# Libica models
from libica.openapi.v3.models import (
    StorageConfigurationWithDetails
)

# Function imports
from .functions.storage_configuration_functions import (
    # Models
    StorageConfigurationObjectModel,
    ProjectToStorageMappingDictModel,
    # Storage Configuration functions
    get_storage_configuration_list,
    set_storage_configuration_list,
    get_project_to_storage_configuration_mapping_list,
    get_project_id_by_s3_key_prefix,
    get_s3_key_prefix_by_project_id,
    convert_s3_uri_to_icav2_uri,
    convert_icav2_uri_to_s3_uri,
    convert_project_data_obj_to_s3_uri,
    convert_s3_uri_to_project_data_obj,
    unpack_s3_uri
)

__all__ = [
    # Libica models
    "StorageConfigurationWithDetails",
    # Internal Models
    "StorageConfigurationObjectModel",
    "ProjectToStorageMappingDictModel",
    # Storage Configuration
    'get_storage_configuration_list',
    'set_storage_configuration_list',
    'get_project_to_storage_configuration_mapping_list',
    'get_project_id_by_s3_key_prefix',
    'get_s3_key_prefix_by_project_id',
    'convert_s3_uri_to_icav2_uri',
    'convert_icav2_uri_to_s3_uri',
    'convert_project_data_obj_to_s3_uri',
    'convert_s3_uri_to_project_data_obj',
    'unpack_s3_uri'
]


#!/usr/bin/env python

# Import libica models
from libica.openapi.v3.models import Project


# Import local functions
from .functions.project_functions import (
    # Project functions
    coerce_project_id_or_name_to_project_obj,
    get_project_obj_from_project_id,
    get_project_obj_from_project_name,
    get_project_id_from_project_name,
    check_project_has_data_sharing_enabled,
    list_projects,
    coerce_project_id_or_name_to_project_id,
    get_project_id,
    get_project_name_from_project_id
)

__all__ = [
    # Libica models
    'Project',
    # Functions
    'coerce_project_id_or_name_to_project_obj',
    'get_project_obj_from_project_id',
    'get_project_obj_from_project_name',
    'get_project_id_from_project_name',
    'check_project_has_data_sharing_enabled',
    'list_projects',
    'coerce_project_id_or_name_to_project_id',
    'get_project_id',
    'get_project_name_from_project_id',
]



#!/usr/bin/env python3

"""
Functions that don't quite do anywhere else
"""
# Standard imports
import re
from typing import Dict, Any, Type, Union
from urllib.parse import urlparse
from uuid import UUID

from pydantic import UUID4


def camel_to_snake_case(camel_case: str) -> str:
    return re.sub(r'(?<!^)(?=[A-Z])', '_', camel_case).lower()


def snake_to_camel_case(snake_case: str) -> str:
    return ''.join(word.title() for word in snake_case.split('_'))


def to_lower_camel_case(snake_str):
    # We capitalize the first letter of each component except the first one
    # with the 'capitalize' method and join them together.
    camel_string = snake_to_camel_case(snake_str)
    return snake_str[0].lower() + camel_string[1:]


def sanitise_dict_keys(input_dict: Dict) -> Dict:
    output_dict = {}
    key: str
    value: Any
    for key, value in input_dict.items():
        output_dict[camel_to_snake_case(key)] = value
    return output_dict


def is_uuid_format(
        project_id: Union[UUID4, str]
) -> bool:
    # First check if it's already a UUID4 object
    if isinstance(project_id, UUID4):
        return True
    # Next check if it's a valid UUID4 string
    try:
        _ = UUID(project_id, version=4)
        return True
    except ValueError:
        return False


def is_uri_format(uri_str: str) -> bool:
    """
    Determine if the string is a valid URI
    :param uri_str:
    :return:
    """
    try:
        url_obj = urlparse(uri_str)
        if url_obj.scheme and url_obj.netloc:
            return True
        else:
            return False
    except ValueError:
        return False


def is_str_type_representation(value: str, type_: Type) -> bool:
    """
    Check if the string 'value' can be represented as the type 'type'

    :param value:
    :param type_:
    :return:
    """

    try:
        type_(value)
        return True
    except ValueError:
        return False


def nextflow_parameter_to_str(parameter: Any) -> str:
    """
    Coerce the parameter to a string representation of the type

    This is easy for ints + floats, for Boolean types we make sure the string is lowercase

    :param parameter:
    :return:
    """
    if type(parameter) == bool:
        return str(parameter).lower()
    return str(parameter)


def coerce_to_uuid4_obj(
        uuid_str: Union[UUID4, str]
) -> UUID4:
    """
    Coerce a string to a UUID4 object

    :param uuid_str:
    :return:
    """
    if isinstance(uuid_str, UUID4):
        return uuid_str
    elif isinstance(uuid_str, str):
        return UUID4(uuid_str)
    raise TypeError(
        f"Expected input type to be str or UUID4, got {type(uuid_str)}"
    )



#!/usr/bin/env python

# Standard imports
from typing import Union

# Import libica models
from libica.openapi.v3.models import (
    AnalysisQueryParameters,
    AnalysisV3,
    AnalysisV4,
    AnalysisInput,
    AnalysisOutput,
    AnalysisOutputList,
    AnalysisStep,
    AnalysisStepLogs,
    CwlAnalysisInputJson,
    CwlAnalysisOutputJson,
    AnalysisStorageV3,
    AnalysisStorageV4
)

AnalysisType = Union[AnalysisV3, AnalysisV4]
AnalysisStorageType = Union[AnalysisStorageV3, AnalysisStorageV4]

# Import everything
from .functions.project_analysis_functions import (
    # Project Analysis functions
    get_project_analysis_inputs,
    get_analysis_input_object_from_analysis_input_code,
    get_outputs_object_from_analysis_id,
    get_analysis_output_object_from_analysis_output_code,
    get_cwl_outputs_json_from_analysis_id,
    get_analysis_obj_from_analysis_id,
    get_analysis_steps,
    get_analysis_log_from_analysis_step,
    write_analysis_step_logs,
    abort_analysis,
    list_analyses,
    get_cwl_analysis_input_json,
    get_cwl_analysis_output_json,
    analysis_step_to_dict,
    coerce_analysis_id_or_user_reference_to_analysis_id,
    get_analysis_obj_from_user_reference,
    coerce_analysis_id_or_user_reference_to_analysis_obj,
    update_analysis_obj,
    add_tag_to_analysis
)

__all__ = [
    # Models
    'AnalysisQueryParameters',
    'AnalysisType',
    'AnalysisV3',
    'AnalysisV4',
    'AnalysisInput',
    'AnalysisOutput',
    'AnalysisOutputList',
    'AnalysisStep',
    'AnalysisStepLogs',
    'CwlAnalysisInputJson',
    'CwlAnalysisOutputJson',
    'AnalysisStorageType',
    'AnalysisStorageV3',
    'AnalysisStorageV4',
    # Functions
    'get_project_analysis_inputs',
    'get_analysis_input_object_from_analysis_input_code',
    'get_outputs_object_from_analysis_id',
    'get_analysis_output_object_from_analysis_output_code',
    'get_cwl_outputs_json_from_analysis_id',
    'get_analysis_obj_from_analysis_id',
    'get_analysis_steps',
    'get_analysis_log_from_analysis_step',
    'write_analysis_step_logs',
    'abort_analysis',
    'list_analyses',
    'get_cwl_analysis_input_json',
    'get_cwl_analysis_output_json',
    'analysis_step_to_dict',
    'coerce_analysis_id_or_user_reference_to_analysis_id',
    'get_analysis_obj_from_user_reference',
    'coerce_analysis_id_or_user_reference_to_analysis_obj',
    'update_analysis_obj',
    'add_tag_to_analysis',
]



#!/usr/bin/env python

"""
Models to easily import
"""
from typing import Union

# Import all models
from libica.openapi.v3.models import *

Analysis = Union[AnalysisV3, AnalysisV4]
CwlAnalysisInput = Union[CwlAnalysisStructuredInput, CwlAnalysisJsonInput]



from .functions.storage_credentials_functions import (
    StorageCredentialMappingModel,
    get_storage_credential_id_from_s3_uri,
    get_relative_path_from_credentials_prefix
)


__all__ = [
    # Models
    "StorageCredentialMappingModel",
    # Functions
    "get_storage_credential_id_from_s3_uri",
    "get_relative_path_from_credentials_prefix",
]


#!/usr/bin/env python

# Standard imports
from pathlib import Path
from typing import List, Optional, Union
import re

# Libica api imports
from libica.openapi.v3 import ApiClient, ApiException
from libica.openapi.v3.api.bundle_api import BundleApi
from libica.openapi.v3.api.bundle_data_api import BundleDataApi
from libica.openapi.v3.api.bundle_pipeline_api import BundlePipelineApi
from libica.openapi.v3.api.project_api import ProjectApi

# Libica model imports
from libica.openapi.v3.models import (
    Data,
    BundleData,
    BundlePagedList,
    BundlePipeline,
    BundlePipelineList,
    CreateBundle,
    Bundle,
    Links
)
from pydantic import UUID4

# Local imports
from ...literals import BundleStatusType
from ...pipelines.functions.pipelines_functions import get_pipeline_obj_from_pipeline_id
from ...utils.configuration import get_icav2_configuration
from ...utils.globals import (
    LIBICAV2_DEFAULT_PAGE_SIZE,
    FILE_DATA_TYPE,
    FOLDER_DATA_TYPE
)

# Set logger
from ...utils.logger import get_logger
from ...utils.miscell import is_uuid_format, coerce_to_uuid4_obj

logger = get_logger()


def generate_empty_bundle(
        bundle_name: str,
        bundle_version: str,
        short_description: str,
        version_comment: str,
        region_id: Optional[Union[UUID4, str]] = None,
        categories: Optional[List[str]] = None,
        pipeline_release_url: Optional[str] = None
) -> Bundle:
    """
    Generate an empty bundle (in DRAFT format) and return the bundle object

    :param bundle_name:  The name of the bundle
    :param bundle_version:  The version of the bundle
    :param short_description:  The description of the bundle
    :param version_comment:  The description of the version
    :param region_id:  The region id of the bundle
    :param categories:
    :param pipeline_release_url:

    :return: The newly created bundle object
    :rtype: `Bundle <https://umccr.github.io/libica/openapi/v3/docs/Bundle/>`_

    :raises: ApiException

    :Examples:

    .. code-block:: python

        :linenos:

        from wrapica.bundle import generate_empty_bundle
        from wrapica.libica_models import Bundle

        # Initialise bundle
        bundle_obj: Bundle = generate_empty_bundle(
            bundle_name="my-first-bundle",
            bundle_description="A quick description of the bundle",
            bundle_version="1.0.0",
            bundle_version_description="First release of my-first-bundle",
            categories=["Test"]
        )

    """
    from ...region import get_default_region

    # Pipeline release url will soon be used when #156 is resolved
    _ = pipeline_release_url

    # Check if region id is set
    if region_id is None:
        region_id = get_default_region().id

    # Check if categories is set
    if categories is None:
        categories = []

    # Configuration needs manual work with custom access token
    with ApiClient(get_icav2_configuration()) as api_client:
        # Create an instance of the API class
        api_instance = BundleApi(api_client)

    create_bundle = CreateBundle(
        name=bundle_name,
        shortDescription=short_description,
        bundleReleaseVersion=bundle_version,
        bundleVersionComment=version_comment,
        regionId=coerce_to_uuid4_obj(region_id),
        bundleStatus="DRAFT",
        categories=categories,
        links=Links(
            links=[
                # Drop link while https://github.com/umccr-illumina/ica_v2/issues/156 is still active
                # Link(
                #     name="GitHub CWL-ICA Release Page",
                #     url=pipeline_release_url
                # )
            ],
            licenses=[],
            homepages=[],
            publications=[]
        ),
        metadataModelId=None
    )

    # example passing only required values which don't have defaults set
    # and optional values
    try:
        # Create a new bundle
        api_response: Bundle = api_instance.create_bundle(create_bundle=create_bundle)
    except ApiException as e:
        logger.error("Exception when calling BundleApi->create_bundle: %s\n" % e)
        raise ApiException

    return api_response


def get_bundle_obj_from_bundle_id(
        bundle_id: Union[UUID4, str]
) -> Bundle:
    """
    Given a bundle_id, return the bundle object

    :param bundle_id:

    :return: The bundle as an object
    :rtype: `Bundle <https://umccr.github.io/libica/openapi/v3/docs/Bundle/>`_

    :raises: ApiException

    :Examples:

    .. code-block:: python
        :linenos:

        from wrapica.bundle import get_bundle_obj_from_bundle_id
        from wrapica.libica_models import Bundle

        # Get bundle object from the bundle id
        bundle_obj: Bundle = get_bundle_obj_from_bundle_id(
            bundle_id='abcdef-1234'
        )

    """

    # Initialise the API client
    with ApiClient(get_icav2_configuration()) as api_client:
        # Create an instance of the API class
        api_instance = BundleApi(api_client)

    try:
        # Get a bundle by ID.
        api_response: Bundle = api_instance.get_bundle(
            str(bundle_id)
        )
    except ApiException as e:
        logger.error("Exception when calling BundleApi->get_bundle_by_id: %s\n" % e)
        raise ApiException

    return api_response


def get_bundle_obj_from_bundle_name(
        bundle_name: str,
        region_id: Optional[Union[UUID4, str]] = None
) -> Bundle:
    """
    Given a bundle name, return the bundle object

    :param bundle_name:  The bundle name
    :param region_id:  The region id of the bundle

    :return: The bundle as an object

    :rtype: `Bundle <https://umccr.github.io/libica/openapi/v3/docs/Bundle/>`_

    :raises: ApiException

    :Examples:

    .. code-block:: python
        :linenos:

        from wrapica.bundle import get_bundle_obj_from_bundle_name
        from wrapica.libica_models import Bundle

        # Get bundle object from the bundle name
        bundle_obj: Bundle = get_bundle_obj_from_bundle_name(
            bundle_name='my-first-bundle'
        )

    """
    bundle_list = filter_bundles(
        bundle_name=bundle_name,
        region_id=region_id
    )

    if len(bundle_list) == 0:
        logger.error(f"Could not get bundle object from {bundle_name}")
        raise IndexError
    if len(bundle_list) > 1:
        logger.error(f"Multiple bundles found with name {bundle_name}")
        raise IndexError

    return bundle_list[0]


def add_pipeline_to_bundle(
        bundle_id: Union[UUID4, str],
        pipeline_id: Union[UUID4, str]
) -> bool:
    """
    Add a pipeline to a bundle.

    :param bundle_id:
    :param pipeline_id:

    :return: True if successful, False otherwise
    :rtype: bool

    :raises: ApiException

    :Examples:

    .. code-block:: python
        :linenos:

        # Wrapica imports
        from wrapica.pipeline import get_pipeline_obj_from_pipeline_id
        from wrapica.bundle import add_pipeline_to_bundle

        # Set vars
        pipeline_code = "my_pipeline_code"
        bundle_name = "my_bundle_name"

        # Get pipeline id from pipeline code
        pipeline_id = get_pipeline_obj_from_pipeline_id(pipeline_code).id

        # Get bundle id from bundle name
        bundle_id = get_bundle_obj_from_bundle_name(bundle_name).id

        # Link pipeline to bundle
        add_pipeline_to_bundle(bundle_id, pipeline_id)
    """

    # Check Pipeline Status
    pipeline_obj = get_pipeline_obj_from_pipeline_id(pipeline_id)
    if not pipeline_obj.status == "RELEASED":
        logger.warning(
            f"Pipeline '{pipeline_id}' is not released. Please release the pipeline before adding it to a bundle")
        return False

    # Get bundle object to confirm it exists
    try:
        _ = get_bundle_obj_from_bundle_id(bundle_id)
    except ApiException as e:
        logger.error(f"Could not get bundle '{bundle_id}', {e}")
        raise ApiException

    # Initialise the API client
    with ApiClient(get_icav2_configuration()) as api_client:
        # Set accept header for adding pipeline to bundle
        api_client.set_default_header(
            header_name="Accept",
            header_value="application/vnd.illumina.v3+json"
        )
        # Create an instance of the API class
        api_instance = BundlePipelineApi(api_client)

    try:
        # Link a pipeline to a bundle.
        api_instance.link_pipeline_to_bundle(
            str(bundle_id),
            str(pipeline_id),
        )
        return True
    except ApiException as e:
        logger.error("Exception when calling BundlePipelineApi->link_pipeline_to_bundle: %s\n" % e)
        raise ApiException


def add_project_data_to_bundle(
    bundle_id: Union[UUID4, str],
    project_id: Union[UUID4, str],
    data_id: Union[UUID4, str]
) -> bool:
    """
    Add project data to a bundle

    :param bundle_id:
    :param project_id:
    :param data_id:

    :return: True if successful, False otherwise

    :rtype: bool

    :raises: ApiException

    :Examples:

    .. code-block:: python
        :linenos:

        # Imports
        from wrapica.bundle import add_project_data_to_bundle
        from wrapica.project_data import convert_icav2_uri_to_data_obj

        # Set vars
        bundle_id = "abcdef-1234"
        data_uri = "icav2://project_id/path/to/data"

        # Get project id and data id from data uri
        project_data_obj = convert_icav2_uri_to_data_obj(data_uri)

        # Add project data to bundle
        add_project_data_to_bundle(bundle_id, project_data_obj.project_id, project_data_obj.data.id)
    """

    # Get Bundle
    bundle_obj = get_bundle_obj_from_bundle_id(bundle_id)

    # Get the data object
    from ...project_data import get_project_data_obj_by_id
    project_data_obj = get_project_data_obj_by_id(project_id, data_id)

    # Confirm data region and bundle region are the same
    if not str(project_data_obj.data.details.region.id) == str(bundle_obj.region.id):
        logger.error(f"Data region '{project_data_obj.data.details.region.code}' and Bundle region '{bundle_obj.region.code}' are not the same")
        return False

    # Initialise the API client
    with ApiClient(get_icav2_configuration()) as api_client:
        # Create an instance of the API class
        api_instance = BundleDataApi(api_client)

    try:
        # Link a data to a bundle.
        api_instance.link_data_to_bundle(
            str(bundle_id),
            str(project_data_obj.data.id)
        )
        return True
    except ApiException as e:
        logger.error("Exception when calling BundleDataApi->link_data_to_bundle: %s\n" % e)
        raise ApiException


def add_data_to_bundle(
        bundle_id: Union[UUID4, str],
        data_id: Union[UUID4, str]
) -> bool:
    """
    Add data to a bundle

    :param bundle_id:
    :param data_id:

    :return: True if successful, False otherwise
    :rtype: bool

    :raises: ApiException

    :Examples:

    .. code-block:: python
        :linenos:

        # Imports
        from wrapica.bundle import add_project_data_to_bundle
        from wrapica.data import convert_icav2_uri_to_data_obj

        # Set vars
        # Use get_bundle_obj_from_bundle_name to get bundle_id if
        # you have the bundle name and not the bundle id
        bundle_id = "abcdef-1234"
        data_uri = "icav2://project_id/path/to/data"

        # Add project data to bundle
        add_data_to_bundle(bundle_id, convert_icav2_uri_to_data_obj(data_uri).id)
    """

    # Get Bundle
    bundle_obj = get_bundle_obj_from_bundle_id(bundle_id)

    # Get the data object
    from ...data import get_data_obj_from_data_id
    data_obj: Data = get_data_obj_from_data_id(data_id)

    # Confirm data region and bundle region are the same
    if not str(data_obj.details.region.id) == str(bundle_obj.region.id):
        logger.error(f"Data region '{data_obj.details.region.id}' and Bundle region '{bundle_obj.region.id}' are not the same")
        return False

    # Initialise the API client
    with ApiClient(get_icav2_configuration()) as api_client:
        # Create an instance of the API class
        api_client.set_default_header(
            header_name="Accept",
            header_value="application/vnd.illumina.v3+json"
        )
        api_instance = BundleDataApi(api_client)

    try:
        # Link a data to a bundle.
        api_instance.link_data_to_bundle(
            str(bundle_id),
            str(data_id)
        )
        return True
    except ApiException as e:
        logger.error("Exception when calling BundleDataApi->link_data_to_bundle: %s\n" % e)
        raise ApiException


def release_bundle(
    bundle_id: Union[UUID4, str]
):
    """
    Release a bundle, converts a bundle status from DRAFT to RELEASED.

    :param bundle_id:

    :raises: ApiException

    :Examples:

    .. code-block:: python
        :linenos:

        # Imports
        from wrapica.bundle import release_bundle

        # Set vars
        bundle_id = "abcdef-1234"

        # Release bundle
        release_bundle(bundle_id)
    """
    with ApiClient(get_icav2_configuration()) as api_client:
        api_client.set_default_header(
            header_name="Accept",
            header_value="application/vnd.illumina.v3+json"
        )
        # Create an instance of the API class
        api_instance = BundleApi(api_client)

    try:
        # release a bundle
        api_instance.release_bundle(str(bundle_id))
    except ApiException as e:
        logger.error("Exception when calling BundleApi->release_bundle: %s\n" % e)
        raise ApiException

    logger.info(f"Successfully released '{bundle_id}'")


def filter_bundles(
    bundle_name: Optional[str] = None,
    project_id: Optional[Union[UUID4, str]] = None,
    region_id: Optional[Union[UUID4, str]] = None,
    status: Optional[BundleStatusType] = None,
    creator_id: Optional[Union[UUID4, str]] = None
) -> Optional[List[Bundle]]:
    """
    Get a list of bundles but filter by name, region id, status, and creator id

    :param bundle_name:  The name of the bundle
    :param project_id: The project id
    :param region_id:  The region id of the bundle
    :param status:  The status of the bundle
    :param creator_id:  The creator id of the bundle

    :return: List of bundles

    :rtype: List[`Bundle <https://umccr.github.io/libica/openapi/v3/docs/Bundle/>`_]

    :raises: ApiException

    :Examples:

    .. code-block:: python
        :linenos:

        # Imports
        from wrapica.bundle import filter_bundles
        from wrapica.libica_models import Bundle
        from wrapica.enums import BundleStatus
        from wrapica.user import get_user_from_user_name

        # Set vars
        user_name = "Alexis Lucattini"

        # Get creator id from user name
        creator_id = get_user_from_user_name(user_name).id

        # Filter bundles
        bundle_list: List[Bundle] = filter_bundles(
            bundle_status=BundleStatus.RELEASED,
            creator_id=creator_id
        )
    """

    if project_id is not None:
        # Use the project/bundle endpoint instead
        bundle_list = list_bundles_in_project(project_id)
    else:
        # Enter a context with an instance of the API client
        with ApiClient(get_icav2_configuration()) as api_client:
            # Create an instance of the API class
            api_instance = BundleApi(api_client)

        # example passing only required values which don't have defaults set
        bundle_list = []
        page_size = LIBICAV2_DEFAULT_PAGE_SIZE
        page_offset = 0

        try:
            while True:
                # Retrieve a bundle.
                api_response: BundlePagedList = api_instance.get_bundles(
                    page_size=str(page_size),
                    page_offset=str(page_offset)
                )
                bundle_list.extend(api_response.items)

                if page_offset + page_size >= api_response.total_item_count:
                    break
                page_offset += page_size

        except ApiException as e:
            logger.error("Exception when calling BundleApi->get_bundle: %s\n" % e)
            raise ApiException

        # Get each bundle by id
        # Must do this manually due to
        # bundle_obj_list = []
        # for bundle in bundle_list:
        #     try:
        #         bundle_obj_list.append(get_bundle_obj_from_bundle_id(bundle.get("id")))
        #     except TypeError:
        #         logger.warning(f"Could not convert into bundle obj from id {bundle.get('id')} ")

    bundle: Bundle
    returned_bundle_list: List[Bundle] = []
    if bundle_name is not None:
        bundle_name_regex = re.compile(bundle_name)
    else:
        bundle_name_regex = None

    for bundle in bundle_list:
        # Check if name matches
        if bundle_name_regex is not None and not bundle_name_regex.fullmatch(bundle.name):
            continue
        # Check region ids match
        if region_id is not None and not str(bundle.region.id) == str(region_id):
            continue
        # Check bundle status match
        if status is not None and not bundle.status == status:
            continue
        # Check creator id
        if creator_id is not None and not str(bundle.owner_id) == str(creator_id):
            continue

        returned_bundle_list.append(bundle)

    return returned_bundle_list


def list_data_in_bundle(
    bundle_id: Union[UUID4, str]
) -> List[BundleData]:
    """
    Given a bundle id, list data in a bundle

    :param bundle_id:

    :return: List of data items
    :rtype: List[`BundleData <https://umccr.github.io/libica/openapi/v3/docs/BundleData/>`_]

    :raises: ApiException

    :Examples:

    .. code-block:: python
        :linenos:

        # Imports
        from wrapica.bundle import list_data_in_bundles

        list_data_in_bundle(bundle_id='abcdef-1234')

    """
    data_items = []

    page_size = LIBICAV2_DEFAULT_PAGE_SIZE
    page_offset = 0

    # Start the loop
    while True:
        # Use the curl api for now
        # example passing only required values which don't have defaults set
        with ApiClient(get_icav2_configuration()) as api_client:
            api_instance = BundleDataApi(api_client)

        try:
            # Retrieve the list of bundle data.
            api_response = api_instance.get_bundle_data(
                bundle_id=str(bundle_id),
                page_size=str(page_size),
                page_offset=str(page_offset)
            )
        except ApiException as e:
            logger.error("Exception when calling BundleDataApi->get_bundle_data: %s\n" % e)
            raise ApiException

        data_items.extend(api_response.items)

        # Check page offset and page size against total item count
        if page_offset + page_size > api_response.total_item_count:
            break
        page_offset += page_size

    return data_items


def filter_bundle_data_to_top_level_only(
        bundle_data: List[BundleData]
) -> List[BundleData]:
    """
    Filter bundle data to top level only (no subdirectories or files underneath folders)

    :param bundle_data: List of linked bundle data items

    :return: List of top level bundle data items
    :rtype: List[`BundleData <https://umccr.github.io/libica/openapi/v3/docs/BundleData/>`_]

    :Examples:

    .. code-block:: python
        :linenos:

        # Imports
        from wrapica.bundle import filter_bundle_data_to_top_level_only, list_data_in_bundle

        # Set vars
        bundle_id = "abcdef-1234"

        # Get all data
        bundle_data = list_data_in_bundle(bundle_id)

        # Filter to top level only
        top_level_bundle_data = filter_bundle_data_to_top_level_only(bundle_data)
    """
    # Find set of project ids
    project_ids: List[UUID4] = list(set(list(map(
        lambda bundle_data_iter: bundle_data_iter.data.details.owning_project_id,
        bundle_data
    ))))

    # Collect bundle data by owning project id
    bundle_data_by_owning_project_id = {
        project_id: list(
            filter(
                lambda bundle_data_iter: str(bundle_data_iter.data.details.owning_project_id) == str(project_id),
                bundle_data
            )
        )
        for project_id in project_ids
    }

    # Find top level folders in bundle data by removing all subdirectories from list
    top_level_bundle_data_by_owning_project_id = {}
    for project_id, project_bundle_data_list in bundle_data_by_owning_project_id.items():
        # Initialise top level bundle data list for this project
        top_level_bundle_data_folder_list = []
        top_level_bundle_data_file_list = []

        all_folders_sorted = sorted(
            filter(
                lambda bundle_data_iter: bundle_data_iter.data.details.data_type == FOLDER_DATA_TYPE,
                project_bundle_data_list
            ),
            key=lambda bundle_data_sort_iter: bundle_data_sort_iter.data.details.path
        )

        # Find top folders only (where folder is not a child of another folder)
        for index_i, bundle_folder_data_iter_i in enumerate(all_folders_sorted):
            # Skip files
            if not bundle_folder_data_iter_i.data.details.data_type == FOLDER_DATA_TYPE:
                continue
            if any(
                map(
                    lambda bundle_folder_data_iter_j:
                    Path(bundle_folder_data_iter_i.data.details.path).is_relative_to(bundle_folder_data_iter_j.data.details.path),
                    all_folders_sorted[:index_i]
                )
            ):
                continue
            top_level_bundle_data_folder_list.append(bundle_folder_data_iter_i)

        # Find top level files in bundle data by only selecting files that are not children of folders
        for bundle_file_data_iter_i in project_bundle_data_list:
            # Skip folders
            if not bundle_file_data_iter_i.data.details.data_type == FILE_DATA_TYPE:
                continue
            if any(
                map(
                    lambda bundle_data_iter_j:
                        Path(bundle_file_data_iter_i.data.details.path).is_relative_to(Path(bundle_data_iter_j.data.details.path)),
                    top_level_bundle_data_folder_list
                )
            ):
                continue
            top_level_bundle_data_file_list.append(bundle_file_data_iter_i)

        # Add to top level bundle data by owning project id
        top_level_bundle_data_by_owning_project_id[project_id] = top_level_bundle_data_folder_list + top_level_bundle_data_file_list

    return list(
        data_iter
        for project_id in top_level_bundle_data_by_owning_project_id.values()
        for data_iter in project_id
    )


def list_pipelines_in_bundle(
    bundle_id: Union[UUID4, str]
) -> List[BundlePipeline]:
    """
    Given a bundle id, list pipelines in a bundle

    :param bundle_id:  The bundle id

    :return: List of pipeline items
    :rtype: List[`BundlePipeline <https://umccr.github.io/libica/openapi/v3/docs/BundlePipeline/>`_]

    :raises: ApiException

    :Examples:

    .. code-block:: python
        :linenos:

        # Imports
        from wrapica.bundle import list_pipelines_in_bundle

        list_pipelines_in_bundle(bundle_id='abcdef-1234')

    """
    # Use the curl api for now
    # No looping for pipelines
    # Enter a context with an instance of the API client
    with ApiClient(get_icav2_configuration()) as api_client:
        # Create an instance of the API class
        api_instance = BundlePipelineApi(api_client)

    # example passing only required values which don't have defaults set
    try:
        # Retrieve a list of bundle pipelines.
        api_response: BundlePipelineList = api_instance.get_bundle_pipelines(str(bundle_id))
    except ApiException as e:
        logger.error("Exception when calling BundlePipelineApi->get_bundle_pipelines: %s\n" % e)
        raise ApiException

    return api_response.items


def list_bundles_in_project(
    project_id: Union[UUID4, str]
) -> List[Bundle]:
    """
    List bundles in a project

    :param project_id: The project id

    :return: List of bundles linked to this project
    :rtype: List[`Bundle <https://umccr.github.io/libica/openapi/v3/docs/Bundle/>`_]

    :raises: ApiException

    :Examples:

    .. code-block:: python
        :linenos:

        # Imports
        from wrapica.bundle import list_bundles_in_project

        list_bundles_in_project(project_id='abcdef-1234')    
    """
    # while True: no iterator for bundles list
    with ApiClient(get_icav2_configuration()) as api_client:
        # Create an instance of the API class
        api_instance = ProjectApi(api_client)

    # example passing only required values which don't have defaults set
    try:
        # Retrieve project bundles.
        api_response = api_instance.get_project_bundles(str(project_id))
    except ApiException as e:
        logger.warning("Exception when calling ProjectApi->get_project_bundles: %s\n" % e)
        raise ApiException

    return list(map(lambda x: x.bundle, api_response.items))


def link_bundle_to_project(
    project_id: Union[UUID4, str],
    bundle_id: Union[UUID4, str]
):
    """
    Link bundle to project

    :param project_id:  The project id
    :param bundle_id:   The bundle id

    :raises: ApiException

    :Examples:

    .. code-block:: python
        :linenos:

        # Imports
        from wrapica.bundle import link_bundle_to_project

        # Set vars
        project_id = "abcdef-1234"
        bundle_id = "abcdef-1234"

        # Link bundle to project
        link_bundle_to_project(project_id, bundle_id)
    """
    # Check bundle list
    existing_bundles: List[Bundle] = list_bundles_in_project(project_id)

    if any(map(
            lambda bundle_obj: str(bundle_obj.id) == str(bundle_id),
            existing_bundles
    )):
        logger.info(f"Bundle {bundle_id} already in project {project_id}")
        return

    # while True: no iterator for bundles list
    with ApiClient(get_icav2_configuration()) as api_client:
        # Create an instance of the API class
        api_instance = ProjectApi(api_client)

    # example passing only required values which don't have defaults set
    try:
        # Link bundle to project
        api_instance.link_project_bundle(
            project_id=str(project_id),
            bundle_id=coerce_to_uuid4_obj(bundle_id)
        )
    except ApiException as e:
        logger.error("Exception when calling ProjectApi->link_project_bundle: %s\n" % e)
        raise ApiException

    logger.info(f"Successfully linked bundle {bundle_id} to project {project_id}")


def unlink_bundle_from_project(
        project_id: Union[UUID4, str],
        bundle_id: Union[UUID4, str],
):
    """
    Remove the bundle from the project

    :param project_id:  The project id
    :param bundle_id:   The bundle id

    :raises: ApiException

    :Examples:

    .. code-block:: python
        :linenos:

        # Imports
        from wrapica.bundle import unlink_bundle_from_project

        # Set vars
        project_id = "abcdef-1234"
        bundle_id = "abcdef-1234"

        # Unlink bundle to project
        unlink_bundle_from_project(project_id, bundle_id)
    """
    # Check bundle list
    existing_bundles: List[Bundle] = list_bundles_in_project(project_id)

    # Map bundle id to bundle object
    if not any(map(
            lambda bundle_iter_: str(bundle_iter_.id) == str(bundle_id),
            existing_bundles
    )):
        logger.error(f"Bundle '{bundle_id}' is not in this project '{project_id}'")
        raise ValueError

    # while True: no iterator for bundles list
    with ApiClient(get_icav2_configuration()) as api_client:
        # DELETE endpoint needs accept header set
        api_client.set_default_header(
            header_name="Accept",
            header_value="application/vnd.illumina.v3+json"
        )
        # Create an instance of the API class
        api_instance = ProjectApi(api_client)

    # example passing only required values which don't have defaults set
    try:
        # Link bundle to project
        api_instance.unlink_project_bundle(
            project_id=project_id,
            bundle_id=coerce_to_uuid4_obj(bundle_id)
        )
    except ApiException as e:
        logger.error("Exception when calling ProjectApi->unlink_project_bundle: %s\n" % e)
        raise ApiException

    logger.info(f"Successfully unlinked bundle {bundle_id} from project {project_id}")


def deprecate_bundle(
        bundle_id: Union[UUID4, str]
):
    """
    Given a bundle id, deprecate the bundle

    :param bundle_id:
    :return:

    :raises: ApiException

    :Examples:

    .. code-block:: python
        :linenos:

        # Imports
        from wrapica.bundle import deprecate_bundle

        # Set vars
        bundle_id = "abcdef-1234"

        # Deprecate the bundle
        deprecate_bundle(bundle_id)
    """
    with ApiClient(get_icav2_configuration()) as api_client:
        api_client.set_default_header(
            header_name="Accept",
            header_value="application/vnd.illumina.v3+json"
        )
        api_instance = BundleApi(api_client)

    try:
        # Deprecate a bundle
        api_instance.deprecate_bundle(str(bundle_id))
    except ApiException as e:
        logger.error("Exception when calling BundleApi->deprecate_bundle: %s\n" % e)
        raise ApiException

    logger.info(f"Successfully deprecated '{bundle_id}'")


def coerce_bundle_id_or_name_to_bundle_obj(
        bundle_id_or_name: str
) -> Bundle:
    """
    Given either a bundle id or bundle name, return the bundle object

    :param bundle_id_or_name:

    :return: The bundle as an object
    :rtype: `Bundle <https://umccr.github.io/libica/openapi/v3/docs/Bundle/>`_
    """
    if is_uuid_format(bundle_id_or_name):
        return get_bundle_obj_from_bundle_id(bundle_id_or_name)
    return get_bundle_obj_from_bundle_name(bundle_id_or_name)


def coerce_bundle_id_or_name_to_bundle_id(
        bundle_id_or_name: str
) -> str:
    """
    Given either a bundle id or bundle name, return the bundle id

    :param bundle_id_or_name:

    :return: The bundle id
    :rtype: str
    """
    if is_uuid_format(bundle_id_or_name):
        return bundle_id_or_name
    return str(get_bundle_obj_from_bundle_name(bundle_id_or_name).id)



#!/usr/bin/env python3

# Standard imports
from time import sleep
from typing import cast, Union

# Libica API imports
from libica.openapi.v3 import ApiClient, ApiException
from libica.openapi.v3.api.job_api import JobApi

# Libica model imports
from libica.openapi.v3.models import Job
from pydantic import UUID4

# Util imports
from ...literals import JobStatusType
from ...utils.configuration import get_icav2_configuration
from ...utils.miscell import coerce_to_uuid4_obj


def get_job(
    job_id: Union[UUID4, str]
) -> Job:
    """
    Get a job (such as a copy job)

    :param job_id:

    :return: The job object
    :rtype: `Job <https://umccr.github.io/libica/openapi/v3/docs/Job/>`_

    """
    # Get the configuration
    configuration = get_icav2_configuration()

    with ApiClient(configuration) as api_client:
        # Create an instance of the API class
        api_instance = JobApi(api_client)

    # example passing only required values which don't have defaults set
    try:
        # Retrieve a job.
        api_response: Job = api_instance.get_job(job_id=coerce_to_uuid4_obj(job_id))
    except ApiException as e:
        raise ApiException("Exception when calling JobApi->get_job: %s\n" % e)

    return api_response


def wait_for_job_completion(
        job_id: Union[UUID4, str],
        raise_on_failure: bool = True
) -> JobStatusType:

    while True:
        # Get the job objects
        job_obj = get_job(job_id)

        # Get the job status
        job_status = cast(JobStatusType, job_obj.status)

        if job_status in ['SUCCEEDED']:
            return job_status
        elif job_status in ['FAILED', 'PARTIALLY_SUCCEEDED', 'STOPPED']:
            if raise_on_failure:
                raise Exception(f"Job {job_id} failed with status {job_status}")
            else:
                return job_status

        sleep(5)




#!/usr/bin/env python3

# Standard library imports
from io import BytesIO
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Optional, List, Union
from zipfile import ZipFile
from cwl_utils.parser import load_document_by_uri
from pydantic import UUID4

# Libica API imports
from libica.openapi.v3 import ApiClient, ApiException
from libica.openapi.v3.api.pipeline_api import PipelineApi

# Libica model imports
from libica.openapi.v3.models import (
    PipelineFile,
    PipelineV3,
    PipelineV4
)

# Local imports
from ...utils.configuration import get_icav2_configuration
from ...utils.cwl_typing_helpers import WorkflowType
from ...utils.logger import get_logger
from ...utils.miscell import is_uuid_format, coerce_to_uuid4_obj

# Logger
logger = get_logger()

# Custom types
PipelineType = Union[PipelineV3, PipelineV4]


def get_pipeline_obj_from_pipeline_id(
    pipeline_id: Union[UUID4, str]
) -> PipelineType:
    """
    Get the pipeline object from the pipeline id

    :param pipeline_id:

    :return: The pipeline object
    :rtype: `Pipeline <https://umccr.github.io/libica/openapi/v3/docs/Pipeline/>`_

    :raises ApiException: If the pipeline is not found

    :Examples:

    .. code-block:: python
        :linenos:

        from pathlib import Path
        from wrapica.pipelines import get_pipeline_obj_from_pipeline_id

        # Use wrapica.pipelines.get_pipeline_obj_from_pipeline_code
        # If you need to convert a pipeline code to a pipeline object

        pipeline_obj: Pipeline = get_pipeline_obj_from_pipeline_id(
            pipeline_id="pipeline_id"
        )
    """
    with ApiClient(get_icav2_configuration()) as api_client:
        # Force the API client to send back the v4 API
        api_client.set_default_header(
            header_name="Accept",
            header_value="application/vnd.illumina.v4+json"
        )
        # Create an instance of the API class
        api_instance = PipelineApi(api_client)

    # example, this endpoint has no required or optional parameters
    try:
        # Retrieve a list of pipelines.
        api_response: PipelineV4 = api_instance.get_pipeline(str(pipeline_id))
    except ApiException as e:
        logger.error("Exception when calling PipelineApi->get_pipelines: %s\n" % e)
        raise ApiException

    return api_response


def get_pipeline_obj_from_pipeline_code(
    pipeline_code: str
) -> PipelineType:
    """
    Get the pipeline object from the pipeline code

    :param pipeline_code:

    :return: The pipeline object
    :rtype: `Pipeline <https://umccr.github.io/libica/openapi/v3/docs/Pipeline/>`_

    :raises ApiException: If the pipeline is not found

    :Examples:

    .. code-block:: python
        :linenos:

        from pathlib import Path
        from wrapica.pipelines import get_pipeline_obj_from_pipeline_code

        # Use wrapica.pipelines.get_pipeline_obj_from_pipeline_id
        # If you need to convert a pipeline id to a pipeline object
        # Since get_pipeline_obj_from_pipeline_code will need to list all pipelines first
        # To find a matching pipeline, if a user has a pipeline id, it is recommended to use
        # get_pipeline_obj_from_pipeline_id instead of this function.

        pipeline_obj: Pipeline = get_pipeline_obj_from_pipeline_code(
            pipeline_code="pipeline_code"
        )

    """

    try:
        return next(
            filter(
                lambda pipeline_iter: pipeline_iter.code == pipeline_code,
                list_all_pipelines(),
            )
        )
    except StopIteration:
        logger.error(f"Pipeline with code {pipeline_code} not found")
        raise StopIteration


def coerce_pipeline_id_or_code_to_pipeline_obj(pipeline_id_or_code: str) -> PipelineType:
    """
    Given either a pipeline id or code, check if the input value is uuid4 format,

    If so, assume it is a pipeline id and collect the object. Otherwise, assume it is a pipeline code and
    call get_pipeline_obj_from_pipeline_code to get the pipeline object

    :param pipeline_id_or_code:
    :return: The pipeline object
    :rtype: `Pipeline <https://umccr.github.io/libica/openapi/v3/docs/Pipeline/>`_

    :raises ValueError: If the pipeline cannot be found

    :Examples:

    .. code-block:: python
        :linenos:

        from wrapica.project_pipelines import coerce_pipeline_id_or_code_to_pipeline_obj

        pipeline_code = "pipeline-123"

        pipeline_obj = coerce_pipeline_id_or_code_to_pipeline_obj(pipeline_code)

    """

    # Check uuid format
    if is_uuid_format(pipeline_id_or_code):
        return get_pipeline_obj_from_pipeline_id(pipeline_id_or_code)
    else:
        return get_pipeline_obj_from_pipeline_code(pipeline_id_or_code)


def coerce_pipeline_id_or_code_to_pipeline_id(pipeline_id_or_code: str) -> Union[UUID4, str]:
    """
    Given either a pipeline id or code, check if the input value is uuid4 format,
    If so, assume it is a pipeline id. Otherwise, assume it is a pipeline code and
    call get_project_pipeline_id_from_pipeline_code to get the pipeline id

    :param pipeline_id_or_code:    The pipeline id or code to retrieve
    :return: The pipeline id       The pipeline id
    :rtype: str

    :raises: ValueError: If the pipeline cannot be found

    :Examples:

    .. code-block:: python
        :linenos:

        from wrapica.project_pipelines import coerce_pipeline_id_or_code_to_pipeline_id

        project_id = "project-123"
        pipeline_code = "pipeline-123"

        pipeline_id = coerce_pipeline_id_or_code_to_pipeline_id(project_id, pipeline_code)
    """

    # Check uuid format
    if is_uuid_format(pipeline_id_or_code):
        return pipeline_id_or_code

    # If not uuid format, assume it is a pipeline code
    return get_pipeline_obj_from_pipeline_code(pipeline_id_or_code).id


def list_all_pipelines() -> List[PipelineType]:
    """
    List all pipelines available to a user through the pipelines/ endpoint

    :return: List of pipelines
    :rtype: List[`Pipeline <https://umccr.github.io/libica/openapi/v3/docs/Pipeline/>`_]

    :raises ApiException: If the pipeline list cannot be retrieved

    :Examples:

    .. code-block:: python
        :linenos:

        from pathlib import Path
        from wrapica.pipelines import list_all_pipelines

        pipelines: List[Pipeline] = list_all_pipelines()
    """

    # Create an instance of the API class
    with ApiClient(get_icav2_configuration()) as api_client:
        # Force the API client to send back the v3 API
        api_client.set_default_header(
            header_name="Accept",
            header_value="application/vnd.illumina.v3+json"
        )
        api_instance = PipelineApi(api_client)

    # example, this endpoint has no required or optional parameters
    # No page token required
    try:
        # Retrieve a list of pipelines.
        pipelines: List[PipelineV3] = api_instance.get_pipelines().items
    except ApiException as e:
        logger.error("Could not get pipeline list")
        raise ApiException from e

    return pipelines


def download_pipeline_file(
    pipeline_id: Union[UUID4, str],
    file_id: Union[UUID4, str],
    file_path: Optional[Path] = None
) -> Optional[BytesIO]:
    """
    Download pipeline file

    :param pipeline_id:  The pipeline id
    :param file_id:  The file id
    :param file_path:  The file path to save the file to (if not set, the file content will be returned)

    :return: The file content if file_path is not set
    :rtype: `BytesIO <https://docs.python.org/3/library/io.html#io.BytesIO>`_

    :raises ApiException: If the file cannot be downloaded

    :Examples:

    .. code-block:: python
        :linenos:

        from pathlib import Path
        from wrapica.pipelines import download_pipeline_file

        download_pipeline_file(
            pipeline_id="pipeline_id",
            file_id="file_id",
            file_path=Path("path/to/file")
        )

    """
    assert file_path.parent.is_dir(), f"Parent directory {file_path.parent} does not exist"

    # Create an instance of the API class
    with ApiClient(get_icav2_configuration()) as api_client:
        # Force the API client to send back the v3 API
        api_client.set_default_header(
            header_name="Accept",
            header_value="application/octet-stream"
        )
        api_instance = PipelineApi(api_client)

    try:
        # Download the contents of a pipeline file.
        api_response = api_instance.download_pipeline_file_content(
            pipeline_id=str(pipeline_id),
            file_id=coerce_to_uuid4_obj(file_id),
        )
    except ApiException as e:
        logger.error("Exception when calling PipelineApi->download_pipeline_file_content: %s\n" % e)
        raise ApiException

    # Write out file
    if file_path is not None:
        # Check parent exists
        assert file_path.parent.is_dir(), f"Parent directory {file_path.parent} does not exist"

        with open(file_path, 'wb') as file_h:
            file_h.write(
                api_response
            )
        return None
    else:
        return BytesIO(api_response)


def list_pipeline_files(
    pipeline_id: Union[UUID4, str]
) -> List[PipelineFile]:
    """
    List pipeline files

    :param pipeline_id:

    :return: List of pipeline files
    :rtype: List[`PipelineFile <https://umccr.github.io/libica/openapi/v3/docs/PipelineFile/>`_]

    :raises ApiException: If the pipeline files cannot be retrieved

    :Examples:

    .. code-block:: python
        :linenos:

        # Imports
        from wrapica.pipelines import list_pipeline_files

        # List pipeline files
        pipeline_files: List[PipelineFile] = list_pipeline_files(
            pipeline_id="pipeline_id"
        )
    """
    # Enter a context with an instance of the API client
    with ApiClient(get_icav2_configuration()) as api_client:
        # Create an instance of the API class
        api_instance = PipelineApi(api_client)

    # example passing only required values which don't have defaults set
    try:
        # Retrieve files for a project pipeline.
        api_response = api_instance.get_pipeline_files(str(pipeline_id))
    except ApiException as e:
        logger.error("Exception when calling ProjectPipelineApi->get_pipeline_files1: %s\n" % e)
        raise ApiException

    return api_response.items


def download_pipeline_to_directory(
        pipeline_id: Union[UUID4, str],
        output_directory: Path
):
    """
    Download a pipeline to a directory

    :param pipeline_id
    :param output_directory

    :raises ApiException: If the pipeline files cannot be retrieved
    :raises AssertionError: If the parent directory does not exist

    :Examples:

    .. code-block:: python
        :linenos:

        # Imports
        from pathlib import Path
        from wrapica.pipelines import download_pipeline_to_directory

        # Download pipeline to local directory
        download_pipeline_to_directory(
            pipeline_id="pipeline_id",
            output_directory=Path("path/to/output/directory")
        )
    """

    # Ensure parent directory exists
    assert output_directory.parent.is_dir(), f"Parent directory {output_directory.parent} does not exist"

    # Create output directory
    output_directory.mkdir(exist_ok=True)

    for pipeline_file in list_pipeline_files(pipeline_id):
        # Get file path
        file_path = output_directory / pipeline_file.name
        # Make sure parent directory exists
        file_path.parent.mkdir(parents=True, exist_ok=True)
        # Download file
        download_pipeline_file(pipeline_id, pipeline_file.id, file_path)


def download_pipeline_to_zip(
        pipeline_id: Union[UUID4, str],
        zip_path: Path
):
    """
    Given a pipeline id, download the pipeline to a zip file

    :param pipeline_id:
    :param zip_path:
    :return:
    """
    # Get the pipeline as an object
    pipeline_obj = get_pipeline_obj_from_pipeline_id(pipeline_id)

    # Get the pipeline code
    pipeline_code = pipeline_obj.code

    # Create the temporary directory
    with TemporaryDirectory() as tmp_dir:
        output_dir = Path(tmp_dir) / pipeline_code

        # Download the pipeline to the directory
        download_pipeline_to_directory(pipeline_id, output_dir)

        # Zip the directory
        tmp_zip_path = output_dir.with_suffix(".zip")

        # Zip the output directory to the zip path
        with ZipFile(tmp_zip_path, 'w') as zip_h:
            for file in output_dir.rglob("*"):
                zip_h.write(file, output_dir.name / file.relative_to(output_dir))

        # Move the zip file to the final location
        tmp_zip_path.rename(zip_path)


def get_cwl_obj_from_pipeline_id(
        pipeline_id: Union[UUID4, str]
) -> WorkflowType:
    """
    Get the pipeline files from a project pipeline
    Arrange the files as they're named to generate the workflow object

    :param pipeline_id:
    :return:
    """

    # Create a temp directory
    pipeline_tmp_dir_obj = TemporaryDirectory(delete=False)
    pipeline_tmp_dir_path = Path(pipeline_tmp_dir_obj.name)

    # Download pipeline to directory
    download_pipeline_to_directory(
        pipeline_id=pipeline_id,
        output_directory=pipeline_tmp_dir_path
    )

    # Get the cwl file
    workflow_file = pipeline_tmp_dir_path / "workflow.cwl"

    # Check the workflow file exists
    if not workflow_file.exists():
        raise FileNotFoundError(f"Expected file 'workflow.cwl' in top directory, but it was not found")

    # Load the document
    return load_document_by_uri(workflow_file)



#!/usr/bin/env python3

# Standard imports
"""
All the enums!
"""
from enum import Enum


class BundleStatus(Enum):
    DRAFT = "DRAFT"
    RELEASED = "RELEASED"
    DEPRECATED = "DEPRECATED"


class Data(Enum):
    FOLDER = "FOLDER"
    FILE = "FILE"


# Deprecated: Use Data instead or literal DataType from literals
DataType = Data


class AnalysisStorageSize(Enum):
    SMALL = "Small"
    MEDIUM = "Medium"
    LARGE = "Large"
    XLARGE = "XLarge"
    XXLARGE = "2XLarge"
    XXXLARGE = "3XLarge"


class PipelineStatus(Enum):
    DRAFT = "DRAFT"
    RELEASED = "RELEASED"


class ProjectDataSortParameter(Enum):
    TIME_CREATED = "timeCreated"
    TIME_CREATED_DESC = "-timeCreated"
    TIME_MODIFIED = "timeModified"
    TIME_MODIFIED_DESC = "-timeModified"
    NAME = "name"
    NAME_DESC = "-name"
    PATH = "path"
    PATH_DESC = "-path"
    FILE_SIZE_IN_BYTES = "fileSizeInBytes"
    FILE_SIZE_IN_BYTES_DESC = "-fileSizeInBytes"
    STATUS = "status"
    STATUS_DESC = "-status"
    FORMAT = "format"
    FORMAT_DESC = "-format"
    DATATYPE = "dataType"
    DATATYPE_DESC = "-dataType"
    WILL_BE_ARCHIVED_AT = "willBeArchivedAt"
    WILL_BE_ARCHIVED_AT_DESC = "-willBeArchivedAt"
    WILL_BE_DELETED_AT = "willBeDeletedAt"
    WILL_BE_DELETED_AT_DESC = "-willBeDeletedAt"


class ProjectDataStatusValues(Enum):
    PARTIAL = "PARTIAL"
    AVAILABLE = "AVAILABLE"
    ARCHIVING = "ARCHIVING"
    ARCHIVED = "ARCHIVED"
    UNARCHIVING = "UNARCHIVING"
    DELETING = "DELETING"


class ProjectAnalysisSortParameters(Enum):
    REFERENCE = "reference"
    REFERENCE_DESC = "-reference"
    USER_REFERENCE = "userReference"
    USER_REFERENCE_DESC = "-userReference"
    PIPELINE = "pipeline"
    PIPELINE_DESC = "-pipeline"
    STATUS = "status"
    STATUS_DESC = "-status"
    START_DATE = "startDate"
    START_DATE_DESC = "-startDate"
    END_DATE = "endDate"
    END_DATE_DESC = "-endDate"
    SUMMARY = "summary"
    SUMMARY_DESC = "-summary"


class ProjectAnalysisStatus(Enum):
    # In process
    REQUESTED = "REQUESTED"
    QUEUED = "QUEUED"
    INITIALIZING = "INITIALIZING"
    PREPARING_INPUTS = "PREPARING_INPUTS"
    IN_PROGRESS = "IN_PROGRESS"
    GENERATING_OUTPUTS = "GENERATING_OUTPUTS"
    # Terminal Successful
    SUCCEEDED = "SUCCEEDED"
    # Terminal Failed
    ABORTING = "ABORTING"
    FAILED = "FAILED"
    FAILED_FINAL = "FAILED_FINAL"
    ABORTED = "ABORTED"
    # Unknown (maybe a nf thing)
    AWAITING_INPUT = "AWAITING_INPUT"


class ProjectAnalysisStepStatus(Enum):
    FAILED = "FAILED"
    DONE = "DONE"
    RUNNING = "RUNNING"
    INTERRUPTED = "INTERRUPTED"
    ABORTED = "ABORTED"
    WAITING = "WAITING"


class WorkflowLanguage(Enum):
    CWL = "CWL"
    NEXTFLOW = "NEXTFLOW"


class StructuredInputParameterType(Enum):
    BOOLEAN = "boolean"
    STRING = "string"
    INTEGER = "integer"
    OPTION = "options"
    FLOAT = "float"


class StructuredInputParameterTypeMapping(Enum):
    BOOLEAN = bool
    STRING = str
    INTEGER = int
    OPTION = str
    FLOAT = float


class AnalysisLogStreamName(Enum):
    STDOUT = "stdout"
    STDERR = "stderr"


class JobStatus(Enum):
    INITIALIZED = "INITIALIZED"
    WAITING_FOR_RESOURCES = "WAITING_FOR_RESOURCES"
    RUNNING = "RUNNING"
    STOPPED = "STOPPED"
    SUCCEEDED = "SUCCEEDED"
    PARTIALLY_SUCCEEDED = "PARTIALLY_SUCCEEDED"
    FAILED = "FAILED"


class Uri(Enum):
    S3 = "s3"
    ICAV2 = "icav2"


class DataTag(Enum):
    TECHNICAL_TAG = "technical_tag"
    USER_TAG = "user_tag"
    CONNECTOR_TAG = "connector_tag"
    RUN_IN_TAG = "run_in_tag"
    RUN_OUT_TAG = "run_out_tag"
    REFERENCE_TAG = "reference_tag"


class AnalysisTag(Enum):
    USER_TAG = "user_tag"
    TECHNICAL_TAG = "technical_tag"
    REFERENCE_TAG = "reference_tag"



from libica.openapi.v3.models import (
    BundleData,
    BundlePagedList,
    BundlePipeline,
    BundlePipelineList,
    CreateBundle,
    Bundle,
    Links
)

from .functions.bundle_functions import (
    generate_empty_bundle,
    get_bundle_obj_from_bundle_id,
    get_bundle_obj_from_bundle_name,
    add_pipeline_to_bundle,
    add_project_data_to_bundle,
    add_data_to_bundle,
    release_bundle,
    filter_bundles,
    list_data_in_bundle,
    filter_bundle_data_to_top_level_only,
    list_pipelines_in_bundle,
    list_bundles_in_project,
    link_bundle_to_project,
    unlink_bundle_from_project,
    deprecate_bundle,
    coerce_bundle_id_or_name_to_bundle_obj,
    coerce_bundle_id_or_name_to_bundle_id
)

__all__ = [
    # models
    'BundleData',
    'BundlePagedList',
    'BundlePipeline',
    'BundlePipelineList',
    'CreateBundle',
    'Bundle',
    'Links',
    # Functions
    'generate_empty_bundle',
    'get_bundle_obj_from_bundle_id',
    'get_bundle_obj_from_bundle_name',
    'add_pipeline_to_bundle',
    'add_project_data_to_bundle',
    'add_data_to_bundle',
    'release_bundle',
    'filter_bundles',
    'list_data_in_bundle',
    'filter_bundle_data_to_top_level_only',
    'list_pipelines_in_bundle',
    'list_bundles_in_project',
    'link_bundle_to_project',
    'unlink_bundle_from_project',
    'deprecate_bundle',
    'coerce_bundle_id_or_name_to_bundle_obj',
    'coerce_bundle_id_or_name_to_bundle_id'
]



# Libica model imports
from libica.openapi.v3.models import (
    Data,
    ProjectData
)


# Local imports
from .functions.data_functions import (
    get_data_obj_from_data_id,
    get_owning_project_id,
    get_project_data_obj_from_data_id,
    convert_icav2_uri_to_data_obj,
    convert_data_obj_to_icav2_uri,
    coerce_data_id_path_or_icav2_uri_to_data_obj
)

__all__ = [
    # Libica model imports
    'Data',
    'ProjectData',
    # Local functions
    'get_data_obj_from_data_id',
    'get_owning_project_id',
    'get_project_data_obj_from_data_id',
    'convert_icav2_uri_to_data_obj',
    'convert_data_obj_to_icav2_uri',
    'coerce_data_id_path_or_icav2_uri_to_data_obj'
]



#!/usr/bin/env python

# External imports
import json
from pathlib import Path
from typing import List, Dict, Union
from xml.dom.minidom import Document
import pandas as pd
from libica.openapi.v3 import ApiClient
from ruamel.yaml import CommentedMap, CommentedSeq
from pydantic import UUID4


# Libica imports
from libica.openapi.v3.models import (
    PipelineLanguageVersion
)
from libica.openapi.v3.api.pipeline_language_api import PipelineLanguageApi

# Wrapica imports
from .configuration import get_icav2_configuration
from .globals import NEXTFLOW_PROCESS_LABEL_RE_OBJ, NEXTFLOW_TASK_POD_MAPPING, DEFAULT_NEXTFLOW_VERSION


def convert_base_config_to_icav2_base_config(base_config_path: Path):
    """
    Given a base config file, overwrite it to an ICA base config file
    By performing the following operations

    withLabel:process_single {
        cpus   = { ... }
        memory = { ... }
        time   = { ... }
    }

    Converted to

    withLabel:process_single {
        cpus   = { ... }
        memory = { ... }
        time   = { ... }
        pod = [ annotation: 'scheduler.illumina.com/presetSize' , value: 'standard-small' ]
    }

    With the following label mapping
    process_single -> standard-small
    process_low -> standard-medium
    process_high -> standard-xlarge
    process_high_memory  -> himem-large

    :param base_config_path:
    :return:
    """
    with open(base_config_path, 'r') as base_config_file:
        base_config = base_config_file.readlines()

    new_base_config = []
    process_label = None
    for line in base_config:
        if NEXTFLOW_PROCESS_LABEL_RE_OBJ.match(line.strip()) is not None:
            process_label = NEXTFLOW_PROCESS_LABEL_RE_OBJ.match(line.strip()).group(1)
        if process_label is not None and line.strip() == '}' and process_label in NEXTFLOW_TASK_POD_MAPPING.keys():
            new_base_config.extend([
                f'        // Added by Wrapica to allow usability icav2\n',
                f'        pod = [ annotation: \'scheduler.illumina.com/presetSize\' , value: \'{NEXTFLOW_TASK_POD_MAPPING[process_label]}\' ]\n',
                line
            ])
            process_label = None
        else:
            new_base_config.append(line)

    # This configuration should also set the base params and docker params
    new_base_config.extend(
        """
// Added by Wrapica to allow usability icav2
params {
    outdir = 'out'  
    publish_dir_mode = 'copy'  
    igenomes_ignore = false  
    enable_conda = false  
    email_on_fail = null  
    email = null  
    ica_smoke_test = false
}

// Edited by Wrapica to allow usability icav2
docker {
    enabled = true
}

        """
    )

    with open(base_config_path, 'w') as new_base_config_file:
        new_base_config_file.writelines(new_base_config)


def write_params_xml_from_nextflow_schema_json(
        nextflow_schema_json_path: Path,
        params_xml_path: Path,
        pipeline_name: str,
        pipeline_version: str
):
    """
    Get the params xml from the nextflow json input file (not the assets/schema_input.json file)
    :param nextflow_schema_json_path:
    :param params_xml_path:
    :param pipeline_name:
    :param pipeline_version:
    :return:
    """

    # Initialise the xml doc
    doc = Document()

    # Create root element
    pipeline = doc.createElement('pd:pipeline')
    pipeline.setAttribute('xmlns:pd', 'xsd://www.illumina.com/ica/cp/pipelinedefinition')
    pipeline.setAttribute('code', pipeline_name)
    pipeline.setAttribute('version', pipeline_version)
    doc.appendChild(pipeline)

    # Create dataInputs element
    data_inputs_element = doc.createElement('pd:dataInputs')
    pipeline.appendChild(data_inputs_element)

    # Create the steps elements
    steps = doc.createElement('pd:steps')
    pipeline.appendChild(steps)

    # Load nextflow_schema_json_path as a dict
    with open(nextflow_schema_json_path, 'r') as nextflow_schema_json_file:
        nextflow_schema_json = json.load(nextflow_schema_json_file)

    # Get the input output options from the nextflow schema json
    # And filter out the generic options
    all_options = dict(
        filter(
            # We don't want the generic options though
            lambda kv: kv[0] not in ["generic_options", "institutional_config_options"],
            nextflow_schema_json.get("definitions").items()
        )
    )

    # Get the params from the input output options
    for option_category_name, option_category_schema_definition in all_options.items():
        property_definitions = option_category_schema_definition.get("properties", {})
        required_property_names = option_category_schema_definition.get("required", [])
        # Create step element
        category_step = doc.createElement('pd:step')
        steps.appendChild(category_step)
        category_step.setAttribute('execution', 'MANDATORY')
        category_step.setAttribute('code', option_category_name)

        # Add label and description
        category_label_element = doc.createElement('pd:label')
        category_label_element.appendChild(doc.createTextNode(option_category_name))
        category_step.appendChild(category_label_element)
        category_description_element = doc.createElement('pd:description')
        if not option_category_schema_definition.get("description", "") == "":
            category_description_element.appendChild(doc.createTextNode(option_category_schema_definition.get("description", "")))
        else:
            category_description_element.appendChild(doc.createTextNode(option_category_name))
        category_step.appendChild(category_description_element)

        # Create the tool element
        tool_element = doc.createElement('pd:tool')
        tool_element.setAttribute('code', option_category_name)
        category_step.appendChild(tool_element)

        for property_name, property_schema_dict in property_definitions.items():
            if property_name == "email":
                # We don't want to send an email at the end of the pipeline
                continue
            if property_name == "outdir":
                # We hardcode this to 'out', so we don't make it available in the params xml
                continue
            if property_name in ["config_profile_name", "config_profile_description"]:
                # Don't need these
                continue
            # Data type
            if "type" not in property_schema_dict.keys():
                continue
            if (
                    property_schema_dict["type"] == "string" and
                    (
                            # Format hint
                            (
                                    property_schema_dict.get("format", None) is not None
                                    and (
                                            property_schema_dict.get("format") == "file-path" or
                                            property_schema_dict.get("format") == "directory-path"
                                    )
                            ) or
                            (
                                    property_schema_dict.get("mimetype", None) is not None
                                    and (
                                            property_schema_dict.get("mimetype") == "text/csv" or
                                            property_schema_dict.get("mimetype") == "text/tsv"
                                    )
                            )
                    )
            ):
                # We have a file object that needs to be added
                # Property schema looks like this:
                # {
                #    "type": "string",
                #    "format": "file-path",
                #    "exists": true,
                #    "schema": "assets/schema_input.json",
                #    "mimetype": "text/csv",
                #    "pattern": "^\\S+\\.tsv$",
                #    "description": "Path to comma-separated file containing information about the samples in the experiment.",
                #    "help_text": "You will need to create a design file with information about the samples in your experiment before running the pipeline. Use this parameter to specify its location. It has to be a comma-separated file with 3 columns, and a header row. See [usage docs](https://nf-co.re/airrflow/usage#samplesheet-input).",
                #    "fa_icon": "fas fa-file-csv"
                # }
                data_input_element = doc.createElement('pd:dataInput')
                data_inputs_element.appendChild(data_input_element)

                data_input_element.setAttribute('code', property_name)
                data_input_element.setAttribute('format', (
                    "UNKNOWN" if (
                            not property_schema_dict.get("mimetype") or
                            property_schema_dict.get("mimetype").split("/")[-1].upper() == "PLAIN"
                    )
                    else property_schema_dict.get("mimetype").split("/")[-1].upper()
                ))
                data_input_element.setAttribute('type', (
                    "FILE"
                    if property_schema_dict.get("mimetype") is not None
                       or property_schema_dict.get("format") == "file-path"
                    else "DIRECTORY"
                ))
                data_input_element.setAttribute('required', (
                    "true"
                    if property_name in required_property_names
                    else "false"
                ))
                # We don't allow multi-values just yet
                data_input_element.setAttribute('multiValue', "false")

                label_element = doc.createElement('pd:label')
                label_element.appendChild(doc.createTextNode(property_name))
                data_input_element.appendChild(label_element)

                description_element = doc.createElement('pd:description')
                description_element.appendChild(doc.createTextNode(property_schema_dict.get("description", "")))
                data_input_element.appendChild(description_element)

            # Parameter type
            else:
                parameter_element = doc.createElement('pd:parameter')
                tool_element.appendChild(parameter_element)

                parameter_element.setAttribute('code', property_name)
                parameter_element.setAttribute('minValues', (
                    "1"
                    if property_name in required_property_names
                    else "0"
                ))
                # We don't allow multi-values just yet
                parameter_element.setAttribute('maxValues', "1")
                # Always USER, would love to get some documentation on this one day
                parameter_element.setAttribute('classification', "USER")

                # Set label and description as childs
                label_element = doc.createElement('pd:label')
                label_element.appendChild(doc.createTextNode(property_name))
                parameter_element.appendChild(label_element)
                description_element = doc.createElement('pd:description')
                description_element.appendChild(doc.createTextNode(property_schema_dict.get("description", "")))
                parameter_element.appendChild(description_element)

                # Add type
                if (
                        property_schema_dict["type"] == "string" and
                        property_schema_dict.get("enum", None) is not None
                ):
                    options_type = doc.createElement("pd:optionsType")
                    parameter_element.appendChild(options_type)

                    for option in property_schema_dict.get("enum"):
                        option_type = doc.createElement("pd:option")
                        option_type.appendChild(doc.createTextNode(str(option)))
                        options_type.appendChild(option_type)
                elif (
                        property_schema_dict["type"] == "string"
                ):
                    string_type = doc.createElement("pd:stringType")
                    parameter_element.appendChild(string_type)
                elif (
                        property_schema_dict["type"] == "boolean"
                ):
                    boolean_type = doc.createElement("pd:booleanType")
                    parameter_element.appendChild(boolean_type)
                elif (
                        property_schema_dict["type"] == "integer"
                ):
                    integer_type = doc.createElement("pd:integerType")
                    parameter_element.appendChild(integer_type)
                elif (
                        property_schema_dict["type"] == "number"
                ):
                    float_type = doc.createElement("pd:floatType")
                    parameter_element.appendChild(float_type)

                # Add default
                if "default" in property_schema_dict.keys():
                    default_element = doc.createElement("pd:value")
                    default_element.appendChild(doc.createTextNode(str(property_schema_dict.get("default"))))
                    parameter_element.appendChild(default_element)

            # Put it all together
            with open(params_xml_path, 'w') as params_xml_file_h:
                params_xml_file_h.write(doc.toprettyxml(indent="  ", encoding="UTF-8", standalone=True).decode())
            #params_xml_file_h.write("\n")


def generate_samplesheet_file_from_input_dict(
        samplesheet_dict: List[Dict],
        schema_input_path: Path,
        samplesheet_path: Path
):
    """
    Generate the samplesheet csv from the samplesheet dict
    :param samplesheet_dict:
    :param schema_input_path:
    :param samplesheet_path:
    :return:
    """
    from ..project_data import convert_icav2_uri_to_project_data_obj, create_download_url

    # Collect the samplesheet list as a dataframe
    samplesheet_df = pd.DataFrame(samplesheet_dict)

    # Write the samplesheet to a csv
    # But note that any icav2 file path should first be presigned before being written to the samplesheet
    with open(schema_input_path, 'r') as schema_input_path_h:
        schema_input_dict = json.load(schema_input_path_h)

    # Get schema input properties
    required_columns = schema_input_dict.get("items", {}).get("required", [])

    # Check all required columns are present in samplesheet_df columns
    for required_column in required_columns:
        if required_column not in samplesheet_df.columns:
            raise ValueError(f"Required column {required_column} not found in samplesheet")

    # Collect all ICAv2 URIs, presign them and write them back to the samplesheet
    icav2_uris_to_presign = []
    for index, row in samplesheet_df.iterrows():
        for column in samplesheet_df.columns:
            if row[column].startswith("icav2://"):
                if ';' in row[column]:
                    # We have multiple URIs
                    icav2_uris_to_presign.extend(row[column].split(';'))
                else:
                    # Presign the URI
                    icav2_uris_to_presign.append(row[column])

    # Presign the URIs
    icav2_uris_to_presign = list(set(icav2_uris_to_presign))

    # Generate project data objects for each URI
    icav2_uris_to_project_data_obj_map = dict(
        map(
            lambda icav2_uri: (icav2_uri, convert_icav2_uri_to_project_data_obj(icav2_uri)),
            icav2_uris_to_presign
        )
    )

    # Specify the presigned url to download
    icav2_uris_to_presigned_url = dict(
        map(
            lambda icav2_uri_kv: (icav2_uri_kv[0], create_download_url(icav2_uri_kv[1].project_id, icav2_uri_kv[1].data.id)),
            icav2_uris_to_project_data_obj_map.items()
        )
    )

    # Replace the icav2 URIs with the presigned URLs
    for index, row in samplesheet_df.iterrows():
        for column in samplesheet_df.columns:
            if row[column].startswith("icav2://"):
                if ";" in row[column]:
                    # We have multiple URIs
                    samplesheet_df.at[index, column] = ";".join(
                        map(
                            lambda uri: icav2_uris_to_presigned_url[uri],
                            row[column].split(';')
                        )
                    )
                else:
                    samplesheet_df.at[index, column] = icav2_uris_to_presigned_url[row[column]]

    # Write the samplesheet to csv
    samplesheet_df.to_csv(samplesheet_path, index=False)


def generate_samplesheet_yaml_template_from_schema_input(
        schema_input_path: Path
) -> CommentedSeq:
    """
    Pull in the samplesheet schema input and generate a samplesheet template
    :param schema_input_path:
    :return:
    """

    # Load the schema input
    with open(schema_input_path, 'r') as schema_input_file:
        schema_input = json.load(schema_input_file)

    # Get the required columns
    required_columns = schema_input.get("items", {}).get("required", [])
    all_properties = schema_input.get("items", {}).get("properties", {})

    # Generate the samplesheet yaml
    samplesheet_yaml = CommentedSeq()

    init_record = CommentedMap()
    for property_name, property_object in all_properties.items():
        init_record[property_name] = ""

        if property_name in required_columns:
            init_record.yaml_add_eol_comment(
                key=property_name,
                comment="Required",
            )
        else:
            init_record.yaml_add_eol_comment(
                key=property_name,
                comment="Optional",
            )

    # Append the one record as an example to the samplesheet yaml
    samplesheet_yaml.append(init_record)

    return samplesheet_yaml


def generate_input_yaml_from_schema_json(
        nextflow_schema_json_path: Path
) -> CommentedMap:
    # Load nextflow_schema_json_path as a dict
    with open(nextflow_schema_json_path, 'r') as nextflow_schema_json_file:
        nextflow_schema_json = json.load(nextflow_schema_json_file)

    # Get the input output options from the nextflow schema json
    # And filter out the generic options
    all_options = dict(
        filter(
            # We don't want the generic options though
            lambda kv: kv[0] not in ["generic_options", "institutional_config_options"],
            nextflow_schema_json.get("definitions").items()
        )
    )

    input_yaml = CommentedMap()

    for option_category_name, option_category_schema_definition in all_options.items():

        property_definitions = option_category_schema_definition.get("properties", {})
        required_property_names = option_category_schema_definition.get("required", [])
        first_property_in_category = True

        for property_name, property_schema_dict in property_definitions.items():
            # Don't add in generic options
            if property_name == "email":
                # We don't want to send an email at the end of the pipeline
                continue
            if property_name == "outdir":
                # We hardcode this to 'out', so we don't make it available in the params xml
                continue
            if property_name in ["config_profile_name", "config_profile_description"]:
                # Don't need these
                continue

            # Make sure that this is a proper schema definition
            if "type" not in property_schema_dict.keys():
                continue

            # Check if property is required
            required_string = "Required " if property_name in required_property_names else "Optional "
            input_yaml[property_name] = f""

            # File type
            if (
                    property_schema_dict["type"] == "string" and
                    property_schema_dict.get("format", None) is not None and
                    property_schema_dict.get("format") == "file-path"
            ):
                # We have a file object that needs to be added
                # Property schema looks like this:
                # {
                #    "type": "string",
                #    "format": "file-path",
                #    "exists": true,
                #    "schema": "assets/schema_input.json",
                #    "mimetype": "text/csv",
                #    "pattern": "^\\S+\\.tsv$",
                #    "description": "Path to comma-separated file containing information about the samples in the experiment.",
                #    "help_text": "You will need to create a design file with information about the samples in your experiment before running the pipeline. Use this parameter to specify its location. It has to be a comma-separated file with 3 columns, and a header row. See [usage docs](https://nf-co.re/airrflow/usage#samplesheet-input).",
                #    "fa_icon": "fas fa-file-csv"
                # }
                if property_schema_dict.get("mimetype", None) is not None:
                    input_yaml.yaml_add_eol_comment(
                        key=property_name,
                        comment=f"icav2//project-id/path/to/file.{property_schema_dict['mimetype'].split('/')[-1]}"
                    )
                else:
                    input_yaml.yaml_add_eol_comment(
                        key=property_name,
                        comment=f": icav2//project-id/path/to/file"
                    )

            # Directory type
            elif (
                    property_schema_dict["type"] == "string" and
                    property_schema_dict.get("format", None) is not None and
                    property_schema_dict.get("format") == "directory-path"
            ):
                input_yaml[property_name] = f""
                input_yaml.yaml_add_eol_comment(
                    key=property_name,
                    comment=f"icav2//project-id/path/to/directory"
                )

            # Enum Type
            elif (
                    property_schema_dict["type"] == "string" and
                    property_schema_dict.get("enum", None) is not None
            ):
                input_yaml.yaml_add_eol_comment(
                    key=property_name,
                    comment=f"One of {' | '.join(map(str, property_schema_dict['enum']))}"
                )

            # Boolean Type
            elif (
                    property_schema_dict["type"] == "boolean"
            ):
                input_yaml.yaml_add_eol_comment(
                    key=property_name,
                    comment="True | False"
                )

            # Integer Type
            elif (
                    property_schema_dict["type"] == "integer"
            ):
                input_yaml.yaml_add_eol_comment(
                    key=property_name,
                    comment="Integer"
                )

            # Number Type
            elif (
                    property_schema_dict["type"] == "number"
            ):
                input_yaml.yaml_add_eol_comment(
                    key=property_name,
                    comment="Number / Float"
                )

            # Add a comment if this is the fist of the property definitions for this category
            if first_property_in_category:
                # Add comment header
                input_yaml.yaml_set_comment_before_after_key(
                    key=property_name,
                    before=(
                            "\n" +
                            option_category_schema_definition.get("description", "") + "\n" +
                            required_string + property_schema_dict.get("description", "") + (
                                f"\nDefault: {property_schema_dict.get('default', '')}"
                                if property_schema_dict.get('default', '') != '' else ""
                            )
                    ),
                    indent=4
                )
                first_property_in_category = False
            else:
                input_yaml.yaml_set_comment_before_after_key(
                    key=property_name,
                    before=required_string + property_schema_dict.get("description", "") + (
                        (
                            f"\nDefault: {property_schema_dict.get('default', '')}"
                            if property_schema_dict.get('default', '') != '' else ""
                        )
                    ),
                    indent=4
                )

    return input_yaml


def download_nextflow_schema_file_from_pipeline_id(
        pipeline_id: Union[UUID4, str],
        schema_json_path: Path
):
    """
    Download the nextflow schema file from the pipeline id

    :param pipeline_id:
    :param schema_json_path:
    :return:
    """
    from ..pipelines import (
        download_pipeline_file, list_pipeline_files
    )

    schema_json_pipeline_file_id = next(
        filter(
            lambda pipeline_file_iter: pipeline_file_iter.get("name") == "nextflow_schema.json",
            list_pipeline_files(pipeline_id)
        )
    )

    download_pipeline_file(
        pipeline_id=pipeline_id,
        file_id=schema_json_pipeline_file_id.get("id"),
        file_path=schema_json_path
    )


def download_nextflow_schema_input_json_from_pipeline_id(
        pipeline_id: Union[UUID4, str],
        schema_input_json_path: Path
):
    """
    Download the nextflow schema file from the pipeline id

    :param pipeline_id:
    :param schema_input_json_path:
    :return:
    """
    from ..pipelines import (
        download_pipeline_file, list_pipeline_files
    )

    schema_json_pipeline_file_id = next(
        filter(
            lambda pipeline_file_iter: pipeline_file_iter.get("name") == "assets/schema_input.json",
            list_pipeline_files(pipeline_id)
        )
    )

    download_pipeline_file(
        pipeline_id=pipeline_id,
        file_id=schema_json_pipeline_file_id.get("id"),
        file_path=schema_input_json_path
    )


def get_nextflow_pipeline_versions_list() -> List[PipelineLanguageVersion]:
    """
    Get the list of nextflow pipeline versions from the ICAv2 API
    :return:
    """
    # Create the API client
    # Enter a context with an instance of the API client
    with ApiClient(get_icav2_configuration()) as api_client:
        # Create an instance of the API class
        pipeline_language_api = PipelineLanguageApi(api_client)

    # Get the nextflow pipeline versions
    return pipeline_language_api.get_nextflow_versions().items


def get_default_nextflow_pipeline_version() -> PipelineLanguageVersion:
    """
    Get the default nextflow pipeline version from the ICAv2 API
    :return:
    """
    # Get the list of nextflow pipeline versions
    nextflow_pipeline_versions: List[PipelineLanguageVersion] = get_nextflow_pipeline_versions_list()

    # Return the default nextflow pipeline version
    return next(
        filter(
            lambda version_obj_iter_: version_obj_iter_.name == DEFAULT_NEXTFLOW_VERSION,
            nextflow_pipeline_versions
        )
    )


def get_default_nextflow_pipeline_version_id() -> str:
    """
    Get the default nextflow pipeline version name from the ICAv2 API
    :return:
    """
    return str(get_default_nextflow_pipeline_version().id)


#!/usr/bin/env python3

"""
Storage Credentials Functions

This module provides functions mainly to resolve URIs to storage credentials.

This is particularly useful for running analyses using the 'externalData' option.
"""
# Imports
from pathlib import Path
from typing import TypedDict, NotRequired, List, Optional, cast, Union
from os import environ
from urllib.parse import urlunparse, urlparse
from pydantic import UUID4

# Libica imports
from libica.openapi.v3 import ApiException, ApiClient, StorageCredentialList
from libica.openapi.v3.api.storage_credentials_api import StorageCredentialsApi
from ruamel.yaml import YAML

# Local imports
from ...utils.configuration import get_icav2_configuration
from ...utils.logger import get_logger
from ...utils.globals import (
    S3_URI_SCHEME,
)

# Get the logger
logger = get_logger()


# Class
class S3UriListModel(TypedDict):
    bucketName: str
    keyPrefix: NotRequired[str]


class StorageCredentialMappingModel(TypedDict):
    id: str
    name: str
    s3UriList: List[S3UriListModel]


STORAGE_CREDENTIAL_OBJECT_LIST: Optional[List[StorageCredentialMappingModel]] = None
STORAGE_CREDENTIAL_OBJECT_LIST_ENV_VAR = "ICAV2_STORAGE_CREDENTIAL_LIST_FILE"


def get_storage_credential_env_list() -> Optional[List[StorageCredentialMappingModel]]:
    if (
            environ.get(STORAGE_CREDENTIAL_OBJECT_LIST_ENV_VAR) is not None and
            Path(environ.get(STORAGE_CREDENTIAL_OBJECT_LIST_ENV_VAR)).is_file()
    ):
        yaml = YAML()

        with open(environ.get(STORAGE_CREDENTIAL_OBJECT_LIST_ENV_VAR)) as file_handle:
            # Load the YAML file
            return list(map(
                lambda item_iter_: cast(
                    StorageCredentialMappingModel,
                    item_iter_
                ),
                yaml.load(file_handle)
            ))
    return None



def get_storage_credential_api_list() -> List[StorageCredentialMappingModel]:
    # Enter a context with an instance of the API client
    with ApiClient(get_icav2_configuration()) as api_client:
        # Create an instance of the API class
        api_instance = StorageCredentialsApi(api_client)

    try:
        # Retrieve a list of the storage credentials
        api_response: StorageCredentialList = api_instance.get_storage_credentials()
    except ApiException as e:
        logger.error("Exception when calling StorageCredentialsApi->get_storage_credentials: %s\n" % e)
        raise ApiException from e

    return list(map(
        lambda item_iter_: (
            cast(
                StorageCredentialMappingModel,
                cast(
                    object,
                    {
                        "id": item_iter_.id,
                        "name": item_iter_.name,
                        # Unfortunately there is no way to link storage configurations to storage credentials at the moment
                        # I have raised the issue with the Illumina team
                        "s3UriList": []
                    }
                )
            )
        ),
        api_response.items
    ))


def set_storage_credential_list():
    global STORAGE_CREDENTIAL_OBJECT_LIST
    if STORAGE_CREDENTIAL_OBJECT_LIST is None:
        STORAGE_CREDENTIAL_OBJECT_LIST = get_storage_credential_env_list()

    if STORAGE_CREDENTIAL_OBJECT_LIST is not None:
        return

    # No environment variable set, so we need to get the list from the API
    # Which takes longer
    STORAGE_CREDENTIAL_OBJECT_LIST = get_storage_credential_api_list()


def get_storage_credential_list() -> List[StorageCredentialMappingModel]:
    if STORAGE_CREDENTIAL_OBJECT_LIST is None:
        set_storage_credential_list()
        return get_storage_credential_list()
    return STORAGE_CREDENTIAL_OBJECT_LIST


def get_storage_credential_id_from_s3_uri(s3_uri: str) -> Optional[str]:
    """
    Get the storage credential id from an s3 uri
    :param s3_uri:
    :return:
    """
    for credential_obj in get_storage_credential_list():
        for s3_uri_list in credential_obj['s3UriList']:
            if (
                    s3_uri.startswith(
                        str(urlunparse((
                                S3_URI_SCHEME,
                                s3_uri_list['bucketName'],
                                s3_uri_list['keyPrefix'],
                                None, None, None
                        )))
                    )
            ):
                return credential_obj['id']

    return None


def get_relative_path_from_credentials_prefix(
        storage_credential_id: Union[UUID4, str],
        s3_uri: str
) -> str:
    """
    Get the appropriate mount path for an analysis given a
    storage credential id and an s3 uri
    :param storage_credential_id:
    :param s3_uri:
    :return:
    """
    # Get credential object
    credential_object = next(filter(
        lambda credential_iter_: credential_iter_['id'] == storage_credential_id,
        get_storage_credential_list()
    ))

    s3_uri_object: S3UriListModel = next(filter(
        lambda item_iter_: (
            item_iter_['bucketName'] == urlparse(s3_uri).netloc
            and
            urlparse(s3_uri).path.lstrip("/").startswith(item_iter_['keyPrefix'])
        ),
        credential_object['s3UriList']
    ))

    # Return s3 uri
    s3_rel_path = Path(
        urlparse(
            s3_uri
        ).path.lstrip("/")
    ).relative_to(
        Path(
            s3_uri_object['keyPrefix']
        )
    )

    # re add the '/' if omitted
    return str(s3_rel_path) + ("/" if s3_uri.endswith("/") else "")



#!/usr/bin/env python

"""
An agnostic Analysis Object

Both CWL Analysis and NextFlow analysis will use this object
as their parent class

"""

# Standard imports
import json
from pathlib import Path
from typing import List, Optional, Union, Dict, Any, cast
from urllib.parse import urlparse
from datetime import datetime, timezone
from uuid import uuid4
from pydantic import UUID4

# Libica imports
from libica.openapi.v3.models import (
    AnalysisV3,
    AnalysisV4,
    AnalysisOutputMapping,
    CreateAnalysisTag,
    CreateCwlAnalysis,
    CreateNextflowAnalysis,
    CwlAnalysisJsonInput,
    CwlAnalysisStructuredInput,
    ProjectData
)

from libica.openapi.v3 import (
    CreateNextflowWithCustomInputAnalysis,
    CreateCwlWithJsonInputAnalysis,
    CwlAnalysisWithJsonInput,
    NextflowAnalysisWithCustomInput,
    CreateAnalysisLogs
)

# Local imports
from ...literals import AnalysisStorageSizeType
from ...utils.miscell import sanitise_dict_keys
from ...utils.logger import get_logger
from ...utils.globals import (
    FOLDER_DATA_TYPE
)

# Set logger
logger = get_logger()

Analysis = Union[AnalysisV3, AnalysisV4]


class ICAv2AnalysisInput:
    """
    Parent class for CWLAnalysisInput, NextflowAnalysisInput
    """
    def __init__(
        self,
    ):
        pass

    def __call__(
            self
    ) -> Union[
        CwlAnalysisWithJsonInput,
        NextflowAnalysisWithCustomInput,
    ]:
        """
        Create the analysis input
        :return:
        """
        self.validate_input()
        return self.create_analysis_input()

    def create_analysis_input(self):
        """
        Implemented in subclass
        :return:
        """
        raise NotImplementedError

    def validate_input(self):
        """
        For structured inputs, one can first import the pipeline id and compare the input values against those
        specified in the pipeline, and compare required against non-required inputs
        For JSON based inputs, one will need to validate against the JSON schema. This is not implemented yet
        as waiting on PR https://github.com/common-workflow-language/cwl-utils/pull/288
        We also want https://github.com/umccr-illumina/ica_v2/issues/162 resolved so we can validate against the top level workflow object.
        :return:
        """
        raise NotImplementedError


class ICAv2PipelineAnalysisTags:
    """
    List of tags
    """
    def __init__(
        self,
        technical_tags: Union[List | Dict],
        user_tags: Union[List | Dict],
        reference_tags: Union[List | Dict]
    ):
        """
        List of tags to use in the pipeline
        :param technical_tags:
        :param user_tags:
        :param reference_tags:
        """

        # Assign
        self.technical_tags = technical_tags
        self.user_tags = user_tags
        self.reference_tags = reference_tags

        self.clean_tags()

    def clean_tags(self):
        # Now clean up
        for tag_type in ["technical_tags", "user_tags", "reference_tags"]:
            if getattr(self, tag_type) is None:
                continue
            if not isinstance(getattr(self, tag_type), (list, dict)):
                raise ValueError(f"{tag_type} must be a list or a dictionary")
            # For each key-value pair in the dictionary, convert it to a string split by '='
            if isinstance(getattr(self, tag_type), dict):
                setattr(
                    self,
                    tag_type,
                    list(
                        map(
                            lambda tag_iter: f"{tag_iter[0]}={tag_iter[1]}",
                            getattr(self, tag_type).items()
                        )
                    )
                )
            else:
                # Ensure that each item in the tag list is a string
                setattr(
                    self,
                    tag_type,
                    list(
                        map(
                            str,
                            getattr(self, tag_type)
                        )
                    )
                )

    def __call__(self) -> CreateAnalysisTag:
        return CreateAnalysisTag(
            technicalTags=self.technical_tags,
            userTags=self.user_tags,
            referenceTags=self.reference_tags
        )

    def combine_tags(self, analysis_tags: 'ICAv2PipelineAnalysisTags') -> 'ICAv2PipelineAnalysisTags':
        """
        Combine tags from another tag object
        :param analysis_tags:
        :return:
        """
        # Combine the tags
        return ICAv2PipelineAnalysisTags(
            technical_tags=self.technical_tags + analysis_tags.technical_tags,
            user_tags=self.user_tags + analysis_tags.user_tags,
            reference_tags=self.reference_tags + analysis_tags.reference_tags
        )

    @classmethod
    def from_dict(cls, tags_dict):
        # Convert camel cases to snake cases
        tags_dict = sanitise_dict_keys(tags_dict)

        return cls(
                    technical_tags=tags_dict.get("technical_tags", []),
                    user_tags=tags_dict.get("user_tags", []),
                    reference_tags=tags_dict.get("reference_tags", [])
        )


class ICAv2EngineParameters:
    """
    The ICAv2 EngineParameters has the following properties
    """

    # Set in subclasses
    workflow_language = None

    def __init__(
            self,
            project_id: Optional[Union[UUID4, str]] = None,
            pipeline_id: Optional[Union[UUID4, str]] = None,
            analysis_input: Optional[Union[CwlAnalysisWithJsonInput, NextflowAnalysisWithCustomInput]] = None,
            analysis_output: Optional[List[AnalysisOutputMapping]] = None,
            logs_output: Optional[CreateAnalysisLogs] = None,
            tags: Optional[ICAv2PipelineAnalysisTags] = None,
            analysis_storage_id: Optional[Union[UUID4, str]] = None,
            analysis_storage_size: Optional[AnalysisStorageSizeType] = None,
    ):
        # Initialise parameters

        # Launch parameters
        self.project_id: Optional[Union[UUID4, str]] = project_id
        self.pipeline_id: Optional[Union[UUID4, str]] = pipeline_id
        self.analysis_storage_id: Optional[Union[UUID4, str]] = analysis_storage_id
        self.analysis_storage_size: Optional[AnalysisStorageSizeType] = analysis_storage_size

        # Output parameters
        self.analysis_output: Optional[List[AnalysisOutputMapping]] = analysis_output
        self.logs_output: Optional[CreateAnalysisLogs] = logs_output
        self.analysis_input: Optional[Union[CwlAnalysisWithJsonInput, NextflowAnalysisWithCustomInput]] = analysis_input

        # Meta parameters
        self.tags: ICAv2PipelineAnalysisTags = tags

        # self.stream_all_files: Optional[bool] = stream_all_files
        # self.stream_all_directories: Optional[bool] = stream_all_directories

        # Set placeholders
        current_utc_time = datetime.now(timezone.utc)
        self.placeholder_dict: Dict = {
            "__DATE_STR__": current_utc_time.strftime("%Y%m%d"),
            "__TIME_STR__": current_utc_time.strftime("%H%M%S"),
            "__UUID8_STR__": uuid4().hex[:8],
            "__UUID16_STR__": uuid4().hex[:16],
        }

    def __call__(self):
        # Assumed that the following have been run
        # set_launch_parameters
        # set_output_parameters
        # set_meta_parameters
        # All functions can be harmlessly called again
        self.set_launch_parameters()
        self.set_output_parameters()
        self.set_meta_parameters()

        # Now check parameters
        self.check_launch_parameters()
        self.check_output_parameters()
        self.check_meta_parameters()

    def set_launch_parameters(
        self,
        project_id: Optional[Union[UUID4, str]] = None,
        pipeline_id: Optional[Union[UUID4, str]] = None,
        analysis_storage_id: Optional[Union[UUID4, str]] = None,
        analysis_storage_size: Optional[AnalysisStorageSizeType] = None,
        analysis_input: Optional[Union[CwlAnalysisJsonInput, CwlAnalysisStructuredInput]] = None
    ):
        # Local import of functions from classes to avoid circular imports
        from ..functions.project_pipelines_functions import (
            get_default_analysis_storage_id_from_project_pipeline,
            get_analysis_storage_id_from_analysis_storage_size
        )
        # Check for project id
        if project_id is not None:
            self.project_id = project_id

        # Check for pipeline id
        if pipeline_id is not None:
            self.pipeline_id = pipeline_id

        # Check for analysis storage id
        if analysis_storage_id is not None:
            self.analysis_storage_id = analysis_storage_id

        # Check for analysis storage size
        if analysis_storage_size is not None:
            self.analysis_storage_size = analysis_storage_size

        # Check for analysis input
        if analysis_input is not None:
            self.analysis_input = analysis_input

        # If analysis storage size is set, but analysis storage id is not set
        if self.analysis_storage_size is not None and self.analysis_storage_id is None:
            self.analysis_storage_id = get_analysis_storage_id_from_analysis_storage_size(
                project_id=self.project_id,
                analysis_storage_size=self.analysis_storage_size
            )

        elif self.analysis_storage_size is None and self.analysis_storage_id is None:
            # Ensure that project_id and pipeline_id are set at this point
            if self.project_id is None or self.pipeline_id is None:
                logger.error("Cannot get analysis storage id without project id and pipeline id")
                raise ValueError
            # Get the default analysis storage id
            self.analysis_storage_id = get_default_analysis_storage_id_from_project_pipeline(
                self.project_id,
                self.pipeline_id
            )

    def check_launch_parameters(self):
        """
        Check the launch parameters have been set
        :return:
        """
        has_error = False
        for parameter in ["project_id", "pipeline_id", "analysis_storage_id"]:
            if self.__getattribute__(parameter) is None:
                logger.error(f"{parameter} has not been set")
                has_error = True
        if has_error:
            raise ValueError

    def set_output_parameters(
        self,
        analysis_output: Optional[List[AnalysisOutputMapping]] = None,
        logs_output: Optional[CreateAnalysisLogs] = None
    ):
        """
        Set analysis output
        :param analysis_output:
        :param logs_output:
        :return:
        """
        if analysis_output is not None:
            self.analysis_output = analysis_output

        if logs_output is not None:
            self.logs_output = logs_output


    def check_output_parameters(self):
        """
        Ensure that at analysis_output is set
        :return:
        """
        # Ensure that at analysis_output is set
        if (
            self.analysis_output is None
        ):
            # Needed to specify one of the inputs
            logger.error("analysis_output is not set")
            raise ValueError

    def set_meta_parameters(
        self,
        tags: Optional[ICAv2PipelineAnalysisTags] = None,
    ):
        # Set tags if tags is empty
        if tags is not None and self.tags is None:
            self.tags = tags

        # Combine tags if both are set
        elif tags is not None and self.tags is not None:
            self.tags = self.tags.combine_tags(tags)

    def check_meta_parameters(self):
        if self.tags is None:
            logger.warning("No tags set for this analysis")

    def update_engine_parameter(self, attribute_name: str, value: Any):
        self.__setattr__(attribute_name, value)

    def populate_placeholders_in_output_path(self, analysis_path: Path):
        """
        Populate placeholders in the output path
        :param analysis_path:
        :return:
        """
        # Import functions locally to avoid circular imports
        from ...utils import fill_placeholder_path
        return fill_placeholder_path(
            analysis_path,
            self.placeholder_dict
        )


class ICAv2PipelineAnalysis:
    """
    A parent analysis class with the following attributes

    * name / user-reference  # The pipeline name
    * pipeline-id  # The pipeline id
    * tags - Tags used on the analysis ICAv2PipelineAnalysisTags class
    * analysisStorageId
    * analysisOutput ( we skip the outputParentFolderId, it doesn't give us much control)
        * This should be a path-based uri
    * analysisInput ( implemented in the subclass )
    """

    def __init__(
        self,
        # Launch parameters
        user_reference: str,
        project_id: Union[UUID4, str],
        pipeline_id: Union[UUID4, str],
        analysis_input: Union[CwlAnalysisWithJsonInput, NextflowAnalysisWithCustomInput],
        analysis_storage_id: Optional[Union[UUID4, str]] = None,
        analysis_storage_size: Optional[AnalysisStorageSizeType] = None,
        analysis_output_uri: Optional[str] = None,
        ica_logs_uri: Optional[str] = None,
        # Meta parameters
        tags: Optional[ICAv2PipelineAnalysisTags] = None
    ):
        """
        Initialise parent analysis class with the following attributes

        * name / user-reference  # The analysis name - Required
        * project-id  # The project id - Required
        * pipeline-id  # The pipeline id - Required
        * tags - Tags used on the analysis ICAv2PipelineAnalysisTags class  - Required
        * analysis_storage_size - The size of the analysis storage - Optional
            * We can figure this out later
            * We can figure this out later
        * analysisOutput ( we skip the outputParentFolderId, it doesn't give us much control)
            * This should be a path-based uri - Required
        * analysis_input ( implemented in the subclass )
        """
        # Launch parameters
        self.user_reference = user_reference
        self.project_id = project_id
        self.pipeline_id = pipeline_id
        self.analysis_output_uri = analysis_output_uri
        self.ica_logs_uri = ica_logs_uri

        # Meta parameters
        self.tags = tags
        self.analysis_storage_id = analysis_storage_id
        self.analysis_storage_size = analysis_storage_size
        self.analysis_input = analysis_input

        self.engine_parameters: Optional[ICAv2EngineParameters] = None

        if self.analysis_output_uri is not None:
            self.analysis_output = cast(List[AnalysisOutputMapping], [self.get_analysis_output_mapping_from_uri()])
        else:
            self.analysis_output = None

        # Place ICA logs output mapping if set
        if self.ica_logs_uri is not None:
            self.logs = cast(CreateAnalysisLogs, self.get_ica_logs_mapping_from_uri())
        else:
            self.logs = None

        # Set the engine parameters
        self.set_engine_parameters()

        # Set the analysis
        self.analysis: Optional[Union[CreateCwlWithJsonInputAnalysis, CreateNextflowWithCustomInputAnalysis]] = None

    def __call__(self, idempotency_key: Optional[str] = None) -> Analysis:
        """
        Fix up the engine parameters, and generate a CWL Analysis Object
        :return:
        """
        # Check engine parameters
        self.check_engine_parameters()

        # Create analysis
        self.analysis = self.create_analysis()

        return self.launch_analysis(idempotency_key=idempotency_key)

    def get_analysis_output_mapping_from_uri(self) -> AnalysisOutputMapping:
        """
        Convert the analysis output to a mapping
        :return:
        """
        from ...project_data import convert_icav2_uri_to_project_data_obj
        # Ensure that the path attribute of analysis_output_uri ends with /
        if not urlparse(self.analysis_output_uri).path.endswith("/"):
            raise ValueError("The analysis output uri must end with a /")

        analysis_output_obj: ProjectData = convert_icav2_uri_to_project_data_obj(
            self.analysis_output_uri,
            create_data_if_not_found=True
        )

        return AnalysisOutputMapping(
            sourcePath="out/",  # Hardcoded, all workflow outputs should be placed in the out folder,
            type=FOLDER_DATA_TYPE,  # Hardcoded, out directory is a folder
            targetProjectId=str(analysis_output_obj.project_id),
            targetPath=analysis_output_obj.data.details.path,
            actionOnExist="OVERWRITE"  # Hardcoded, always overwrite existing data
        )

    def get_ica_logs_mapping_from_uri(self) -> CreateAnalysisLogs:
        from ...project_data import convert_icav2_uri_to_project_data_obj
        # Ensure that the path attribute of analysis_output_uri ends with /
        if not urlparse(self.ica_logs_uri).path.endswith("/"):
            raise ValueError("The analysis output uri must end with a /")

        ica_logs_project_data_obj: ProjectData = convert_icav2_uri_to_project_data_obj(
            self.ica_logs_uri,
            create_data_if_not_found=True
        )

        return CreateAnalysisLogs(
            logsFolderId=ica_logs_project_data_obj.data.id,
            logsFolderPath=None
        )

    def set_engine_parameters(self):
        """
        Implemented in subclass
        :return:
        """
        raise NotImplementedError

    def check_engine_parameters(self):
        """
        Implemented in subclass
        :return:
        """
        self.engine_parameters()

    def create_analysis(self) -> Union[CreateCwlAnalysis, CreateNextflowAnalysis]:
        # Implemented in subclass
        raise NotImplementedError

    def launch_analysis(self, idempotency_key: Optional[str] = None) -> Analysis:
        # Implemented in subclass
        raise NotImplementedError

    def get_analysis_launch_json(self) -> Dict:
        """
        Save the analysis launch json
        :return:
        """
        return self.analysis.to_dict()

    def save_analysis(self, json_path: Path):
        """
        Save the analysis to a json file
        :param json_path:
        :return:
        """
        analysis_dict = self.get_analysis_launch_json()

        with open(json_path, "w") as f:
            f.write(
                json.dumps(
                    analysis_dict,
                    indent=2
                )
            )



#!/usr/bin/env python3

"""
Nextflow analysis
"""
# Standard imports
import json
from typing import List, Dict, Optional, Union
from pydantic import UUID4

# Libica imports
from libica.openapi.v3 import (
    AnalysisInputDataMount, AnalysisInputExternalData, NextflowAnalysisWithCustomInput,
    AnalysisOutputMapping, CreateNextflowWithCustomInputAnalysis,
    AnalysisV4 as Analysis, CreateAnalysisLogs
)

# Local parent imports
from .analysis import (
    ICAv2PipelineAnalysisTags,
    ICAv2PipelineAnalysis,
    ICAv2EngineParameters,
    ICAv2AnalysisInput
)

# Local imports
from ...literals import (
    WorkflowLanguageType, AnalysisStorageSizeType,
)
from ...utils.logger import get_logger


# Set logger
logger = get_logger()


class ICAv2NextflowAnalysisInput(ICAv2AnalysisInput):
    """
    Generate a Nextflow Analysis input from a JSON input - 
    if the input is an icav2 uri, it is considered an analysis data input
    """

    def __init__(
        self,
        input_json: Dict,
        cache_uri: Optional[str] = None,
    ):
        # Initialise parent class
        super().__init__()

        # Get the cache uri from the input_json if it exists
        # Required if the nextflow pipeline has a parameter called 'input' that is a list of objects
        self.cache_uri: Optional[str] = cache_uri

        # Set inputs
        self.input_json: Dict = input_json

        # Initialise values after dereferencing
        self.data_ids: List[str] = []
        self.input_json_deferenced_dict: Dict = {}
        self.mount_paths_list: List[AnalysisInputDataMount] = []
        self.external_mounts_list: List[AnalysisInputExternalData] = []
        self.input_json_str: Optional[str] = None

    def validate_input(self):
        """
        Cannot yet validate the input as we dont have an easy way to collect a JSON schema for the Nextflow pipeline.
        We can do this for non-proprietary pipelines, but not for proprietary pipelines.
        User is expected to know the input format of the Nextflow pipeline they are using.
        :return:
        """
        pass

    def create_analysis_input(self) -> NextflowAnalysisWithCustomInput:
        # Deference cwl input json
        self.deference_nextflow_input_json()

        # Set input json to string
        self.set_input_json()

        self.validate_input()

        # Generate a CWL analysis input
        return NextflowAnalysisWithCustomInput(
            customInput=self.input_json_str,
            dataIds=self.data_ids,
            mounts=self.mount_paths_list,
            externalData=self.external_mounts_list
        )

    def deference_nextflow_input_json(self):
        from ..functions.project_pipelines_functions import (
            convert_uris_to_data_ids_from_nextflow_input_json
        )
        self.input_json_deferenced_dict, self.mount_paths_list, self.external_mounts_list = convert_uris_to_data_ids_from_nextflow_input_json(
              self.input_json,
              cache_uri=self.cache_uri
        )
        self.data_ids = list(
            map(
                lambda mount_path_iter: mount_path_iter.data_id,
                self.mount_paths_list
            )
        )

    def set_input_json(self):
        if not 'outdir' in self.input_json_deferenced_dict.keys():
            self.input_json_deferenced_dict['outdir'] = 'out/'

        self.input_json_str = json.dumps(self.input_json_deferenced_dict, indent=2)


class ICAv2NextflowEngineParameters(ICAv2EngineParameters):
    """
    The ICAv2 EngineParameters has the following properties
    """

    workflow_language: WorkflowLanguageType = "NEXTFLOW"

    def __init__(
        self,
        project_id: Optional[Union[UUID4, str]] = None,
        pipeline_id: Optional[Union[UUID4, str]] = None,
        analysis_output: Optional[List[AnalysisOutputMapping]] = None,
        logs_output: Optional[CreateAnalysisLogs] = None,
        analysis_input: Optional[NextflowAnalysisWithCustomInput] = None,
        tags: Optional[ICAv2PipelineAnalysisTags] = None,
        analysis_storage_id: Optional[Union[UUID4, str]] = None,
        analysis_storage_size: Optional[AnalysisStorageSizeType] = None,
    ):
        # Initialise parameters
        super().__init__(
            project_id=project_id,
            pipeline_id=pipeline_id,
            analysis_output=analysis_output,
            logs_output=logs_output,
            analysis_input=analysis_input,
            tags=tags,
            analysis_storage_id=analysis_storage_id,
            analysis_storage_size=analysis_storage_size,
        )


class ICAv2NextflowPipelineAnalysis(ICAv2PipelineAnalysis):
    """
    The ICAv2CWLPipelineAnalysis has the following properties
        * user_reference: str
        * input_json: Dict  (cwl_inputs)
        * engine_parameters: ICAv2EngineParameters

    The engineParameters then populate the following parameters for its parent class

    pipeline_id: str,
    tags: ICAv2PipelineAnalysisTags,
    analysis_input: Union[CwlAnalysisJsonInput, NextflowAnalysisInput],
    analysis_output: str,
    analysis_storage_size: Optional[AnalysisStorageSize] = None

    """

    def __init__(
        self,
        # Launch parameters
        user_reference: str,
        project_id: Union[UUID4, str],
        pipeline_id: Union[UUID4, str],
        analysis_input: NextflowAnalysisWithCustomInput,
        analysis_storage_id: Optional[Union[UUID4, str]] = None,
        analysis_storage_size: Optional[AnalysisStorageSizeType] = None,
        # Output parameters
        analysis_output_uri: Optional[str] = None,
        ica_logs_uri: Optional[str] = None,
        # Meta parameters
        tags: Optional[ICAv2PipelineAnalysisTags] = None,
    ):
        """
        Initialise input
        :param user_reference
        :param project_id
        :param pipeline_id
        :param analysis_input
        :param analysis_storage_id
        :param analysis_storage_size
        :param analysis_output_uri
        :param ica_logs_uri
        :param tags
        """
        # Set under parent init script through set_engine_parameters
        self.engine_parameters: Optional[ICAv2NextflowEngineParameters] = None

        # Call parent class
        super().__init__(
            user_reference=user_reference,
            project_id=project_id,
            pipeline_id=pipeline_id,
            analysis_input=analysis_input,
            analysis_storage_id=analysis_storage_id,
            analysis_storage_size=analysis_storage_size,
            analysis_output_uri=analysis_output_uri,
            ica_logs_uri=ica_logs_uri,
            tags=tags
        )

    def set_engine_parameters(self):
        self.engine_parameters = ICAv2NextflowEngineParameters(
            project_id=self.project_id,
            pipeline_id=self.pipeline_id,
            analysis_output=self.analysis_output,
            logs_output=self.logs,
            analysis_input=self.analysis_input,
            tags=self.tags,
            analysis_storage_id=self.analysis_storage_id,
            analysis_storage_size=self.analysis_storage_size,
        )

    def create_analysis(self) -> CreateNextflowWithCustomInputAnalysis:
        return CreateNextflowWithCustomInputAnalysis(
            userReference=self.user_reference,
            pipelineId=self.pipeline_id,
            tags=self.engine_parameters.tags(),
            analysisInput=self.analysis_input,
            analysisStorageId=self.engine_parameters.analysis_storage_id,
            analysisOutput=self.engine_parameters.analysis_output,
            logs=self.engine_parameters.logs_output,
            outputParentFolderId=None,
        )

    def launch_analysis(self, idempotency_key: Optional[str] = None) -> Analysis:
        from ..functions.project_pipelines_functions import launch_nextflow_workflow
        return launch_nextflow_workflow(
            project_id=self.project_id,
            nextflow_analysis=self.analysis,
            idempotency_key=idempotency_key
        )



#!/usr/bin/env python3
# Standard imports
from os import environ
from pathlib import Path
from typing import List, Optional, Tuple, TypedDict, NotRequired, cast, Union
from urllib.parse import urlunparse, urlparse

from pydantic import UUID4
from ruamel.yaml import YAML

# Libica imports
from libica.openapi.v3 import ApiException, ApiClient
from libica.openapi.v3.api.storage_configuration_api import StorageConfigurationApi
from libica.openapi.v3.models import (
    ProjectData,
)

# Local imports
from ...utils.configuration import get_icav2_configuration
from ...utils.logger import get_logger
from ...utils.globals import (
    S3_URI_SCHEME,
    ICAV2_URI_SCHEME,
    FOLDER_DATA_TYPE
)

# Get the logger
logger = get_logger()


class StorageConfigurationObjectModel(TypedDict):
    id: str
    bucketName: str
    keyPrefix: str


class ProjectToStorageMappingDictModel(TypedDict):
    id: str
    name: str
    storageConfigurationId: str
    prefix: NotRequired[str]


STORAGE_CONFIGURATION_OBJECT_LIST: Optional[List[StorageConfigurationObjectModel]] = None
STORAGE_CONFIGURATION_OBJECT_LIST_ENV_VAR = "ICAV2_STORAGE_CONFIGURATION_LIST_FILE"

PROJECT_TO_STORAGE_CONFIGURATION_MAPPING_LIST: Optional[List[ProjectToStorageMappingDictModel]] = None
PROJECT_TO_STORAGE_CONFIGURATION_MAPPING_LIST_ENV_VAR = "ICAV2_PROJECT_TO_STORAGE_CONFIGURATION_MAPPING_LIST_FILE"


def get_storage_configuration_env_list() -> Optional[List[StorageConfigurationObjectModel]]:
    if (
            environ.get(STORAGE_CONFIGURATION_OBJECT_LIST_ENV_VAR) is not None and
            Path(environ.get(STORAGE_CONFIGURATION_OBJECT_LIST_ENV_VAR)).is_file()
    ):
        yaml = YAML()

        with open(environ.get(STORAGE_CONFIGURATION_OBJECT_LIST_ENV_VAR)) as file_handle:
            # Load the YAML file
            return list(map(
                lambda item_iter_: cast(
                    StorageConfigurationObjectModel,
                    item_iter_
                ),
                yaml.load(file_handle)
            ))
    return None


def get_storage_configuration_api_list() -> List[StorageConfigurationObjectModel]:
    # Enter a context with an instance of the API client
    with ApiClient(get_icav2_configuration()) as api_client:
        # Create an instance of the API class
        api_instance = StorageConfigurationApi(api_client)

    # example, this endpoint has no required or optional parameters
    try:
        # Retrieve a list of storage configurations.
        api_response = api_instance.get_storage_configurations()
    except ApiException as e:
        logger.error("Exception when calling StorageConfigurationApi->get_storage_configurations: %s\n" % e)
        raise ApiException from e

    return list(map(
        lambda item_iter_: (
            cast(
                StorageConfigurationObjectModel,
                cast(
                    object,
                    {
                        "id": item_iter_.id,
                        "bucketName": f"s3://{item_iter_.storage_configuration_details.aws_s3.bucket_name}",
                        "keyPrefix": str(Path(item_iter_.storage_configuration_details.aws_s3.key_prefix)) + "/"
                    }
                )
            )
        ),
        api_response.items
    ))


def set_storage_configuration_list():
    global STORAGE_CONFIGURATION_OBJECT_LIST
    if STORAGE_CONFIGURATION_OBJECT_LIST is None:
        STORAGE_CONFIGURATION_OBJECT_LIST = get_storage_configuration_env_list()

    if STORAGE_CONFIGURATION_OBJECT_LIST is not None:
        return

    # No environment variable set, so we need to get the list from the API
    # Which takes longer
    STORAGE_CONFIGURATION_OBJECT_LIST = get_storage_configuration_api_list()


def get_storage_configuration_list() -> List[StorageConfigurationObjectModel]:
    if STORAGE_CONFIGURATION_OBJECT_LIST is None:
        set_storage_configuration_list()
        return get_storage_configuration_list()
    return STORAGE_CONFIGURATION_OBJECT_LIST


def get_project_to_storage_env_mapping() -> Optional[List[ProjectToStorageMappingDictModel]]:
    if (
            environ.get(PROJECT_TO_STORAGE_CONFIGURATION_MAPPING_LIST_ENV_VAR) is not None and
            Path(environ.get(PROJECT_TO_STORAGE_CONFIGURATION_MAPPING_LIST_ENV_VAR)).is_file()
    ):
        yaml = YAML()

        with open(environ.get(PROJECT_TO_STORAGE_CONFIGURATION_MAPPING_LIST_ENV_VAR)) as file_handle:
            # Load the YAML file
            return list(map(
                lambda item_iter_: cast(
                    ProjectToStorageMappingDictModel,
                    item_iter_
                ),
                yaml.load(file_handle)
            ))
    return None


def get_project_to_storage_configuration_api_mapping() -> List[ProjectToStorageMappingDictModel]:
    from wrapica.project import list_projects

    # Get the storage configuration list
    storage_configuration_list: List[StorageConfigurationObjectModel] = get_storage_configuration_list()

    # If there are no storage configurations, return an empty list
    if not storage_configuration_list:
        return []

    # Map the storage configuration id to each project
    return list(map(
      lambda project_iter: (
          cast(
              ProjectToStorageMappingDictModel,
              cast(
                  object,
                  {
                      "id": project_iter.id,
                      "name": project_iter.name,
                      "storageConfigurationId": project_iter.self_managed_storage_configuration.id,
                      "prefix": project_iter.name,
                  }
              )
          )
      ),
      list(filter(
          lambda project_iter: (
              project_iter.self_managed_storage_configuration is not None
          ),
          list_projects()
      ))
    ))


def set_project_to_storage_configuration_mapping():
    global PROJECT_TO_STORAGE_CONFIGURATION_MAPPING_LIST
    if PROJECT_TO_STORAGE_CONFIGURATION_MAPPING_LIST is None:
        PROJECT_TO_STORAGE_CONFIGURATION_MAPPING_LIST = get_project_to_storage_env_mapping()

    if PROJECT_TO_STORAGE_CONFIGURATION_MAPPING_LIST is not None:
        return

    # No environment variable set, so we need to get the list from the API
    # Which takes longer
    PROJECT_TO_STORAGE_CONFIGURATION_MAPPING_LIST = get_project_to_storage_configuration_api_mapping()


def get_project_to_storage_configuration_mapping_list() -> List[ProjectToStorageMappingDictModel]:
    if PROJECT_TO_STORAGE_CONFIGURATION_MAPPING_LIST is None:
        set_project_to_storage_configuration_mapping()
        return get_project_to_storage_configuration_mapping_list()
    return PROJECT_TO_STORAGE_CONFIGURATION_MAPPING_LIST


def get_project_id_by_s3_key_prefix(s3_key_prefix: str) -> Optional[str]:
    # Iterate through the storage configuration mapping dict
    for project_model in get_project_to_storage_configuration_mapping_list():

        project_model_prefix = (
            project_model.get('prefix', None)
            if project_model.get('prefix', None) is not None
            else ""
        )

        # Configuration model
        try:
            configuration_model = next(
                filter(
                    lambda config_iter: (
                        config_iter['id'] == project_model['storageConfigurationId']
                    ),
                    get_storage_configuration_list()
                )
            )
        except StopIteration:
            continue

        # Join the storage configuration prefix with the project prefix
        project_s3_key_prefix = str(urlunparse((
            S3_URI_SCHEME,
            configuration_model['bucketName'],
            str(Path(configuration_model['keyPrefix']) / project_model_prefix) + "/",
            None, None, None
        )))

        if s3_key_prefix.startswith(project_s3_key_prefix):
            return project_model['id']

    logger.error(f"Could not find project id for s3 key prefix: {s3_key_prefix}")
    return None


# And vice-versa
def get_s3_key_prefix_by_project_id(
        project_id: Union[UUID4, str]
) -> Optional[str]:
    # Local imports
    # Return Key Prefix with project name extension
    project_model = next(filter(
        lambda project_iter: (
            project_iter['id'] == project_id
        ),
        get_project_to_storage_configuration_mapping_list()
    ))

    project_model_prefix = (
        project_model.get('prefix', None)
        if project_model.get('prefix', None) is not None
        else ""
    )

    # Configuration model
    configuration_model = next(
        filter(
            lambda config_iter: (
                    config_iter['id'] == project_model['storageConfigurationId']
            ),
            get_storage_configuration_list()
        )
    )

    # Join the storage configuration prefix with the project prefix
    return str(urlunparse((
        S3_URI_SCHEME,
        configuration_model['bucketName'],
        str(
            Path(configuration_model['keyPrefix']) /
            project_model_prefix
        ) + "/",
        None, None, None
    )))


def convert_s3_uri_to_icav2_uri(s3_uri: str) -> str:
    # Convert S3 URI to ICAv2 URI
    # Putting it all together
    # Determine which project id we are working with
    project_id = get_project_id_by_s3_key_prefix(s3_uri)
    # Then remap to determine the root prefix of this project
    project_s3_prefix = get_s3_key_prefix_by_project_id(project_id)

    return str(
        urlunparse(
            (
                ICAV2_URI_SCHEME,
                get_project_id_by_s3_key_prefix(s3_uri),
                # This path is then the relative path to the project s3 prefix
                str(Path(urlparse(s3_uri).path).relative_to(Path(urlparse(project_s3_prefix).path))) + ("/" if s3_uri.endswith("/") else ""),
                None, None, None
            )
        )
    )


def convert_icav2_uri_to_s3_uri(icav2_uri: str) -> str:
    # Convert ICAv2 URI to S3 URI
    # Putting it all together, in-fact this is a little more straight forward
    # Than the inverse
    # Determine which project id we are working with
    from ...project import coerce_project_id_or_name_to_project_id

    # Get the project id from the icav2 uri netloc
    project_id = coerce_project_id_or_name_to_project_id(urlparse(icav2_uri).netloc)

    # Get the s3 key prefix for this project
    project_s3_prefix = get_s3_key_prefix_by_project_id(project_id)

    return str(
        urlunparse(
            (
                urlparse(project_s3_prefix).scheme,
                urlparse(project_s3_prefix).netloc,
                str(
                    Path(urlparse(project_s3_prefix).path).joinpath(
                        # Cannot join two abs paths so we need to strip the leading /
                        # We can then join the relative path to the project s3 prefix
                        Path(str(Path(urlparse(icav2_uri).path)).lstrip("/"))
                    )
                ) + ("/" if icav2_uri.endswith("/") else ""),
                None, None, None
            )
        )
    )


def convert_project_data_obj_to_s3_uri(project_data_obj: ProjectData) -> str:
    # Convert ProjectData object to S3 URI
    project_s3_prefix = get_s3_key_prefix_by_project_id(project_data_obj.data.details.owning_project_id)

    return str(
        urlunparse(
            (
                urlparse(project_s3_prefix).scheme,
                urlparse(project_s3_prefix).netloc,
                str(Path(urlparse(project_s3_prefix).path) / Path(project_data_obj.data.details.path.lstrip("/"))) + ("/" if project_data_obj.data.details.data_type == FOLDER_DATA_TYPE else ""),
                None, None, None
            )
        )
    )


# And vice-versa
def convert_s3_uri_to_project_data_obj(s3_uri: str, create_data_if_not_found: bool = False) -> ProjectData:
    # Convert S3 URI to ProjectData object
    # Easiest to convert to icav2 uri first and then to project data object
    from ...project_data import convert_icav2_uri_to_project_data_obj
    return convert_icav2_uri_to_project_data_obj(
        convert_s3_uri_to_icav2_uri(s3_uri),
        create_data_if_not_found=create_data_if_not_found
    )


def unpack_s3_uri(uri: str) -> Tuple[str, str]:
    # Unpack S3 URI
    # Get the project id from the s3 uri
    from ...project_data import unpack_uri
    return unpack_uri(convert_s3_uri_to_icav2_uri(uri))



#!/usr/bin/env python3

"""
Functions for project management
"""

# Standard library imports
from typing import List, Union
from pydantic import UUID4

# Libica imports
from libica.openapi.v3 import ApiClient, ApiException
from libica.openapi.v3.api.project_api import ProjectApi

# Libica models
from libica.openapi.v3.models import Project

# Local imports
from ...utils.configuration import (
    get_icav2_configuration,
    get_project_id_from_env_var,
    get_project_id_from_session_file
)
from ...utils.logger import get_logger
from ...utils.globals import LIBICAV2_DEFAULT_PAGE_SIZE
from ...utils.miscell import is_uuid_format

# Logger helpers
logger = get_logger()

# GLOBALS
PROJECT_MAPPING_DICT = None


def _set_project_mapping_dict():
    global PROJECT_MAPPING_DICT

    PROJECT_MAPPING_DICT = dict(
        map(
            lambda lambda_project_obj: (lambda_project_obj.id, lambda_project_obj.name),
            list_projects()
        )
    )


def _get_project_mapping_dict():
    if PROJECT_MAPPING_DICT is not None:
        return PROJECT_MAPPING_DICT

    _set_project_mapping_dict()

    return _get_project_mapping_dict()


def get_project_obj_from_project_id(
    project_id: Union[UUID4, str]
) -> Project:
    """
    Given a project id return the project object

    :param project_id: The id of the project

    :return: The project object
    :rtype: List[`Project <https://umccr.github.io/libica/openapi/v3/docs/Project/>`_]

    """

    with ApiClient(get_icav2_configuration()) as api_client:
        api_instance = ProjectApi(api_client)

    try:
        api_response: Project = api_instance.get_project(str(project_id))
    except ApiException as e:
        logger.error("Exception when calling ProjectApi->get_project_by_id: %s\n" % e)
        raise ApiException

    return api_response


def get_project_obj_from_project_name(
    project_name: str
) -> Project:
    """
    Given a project name, get the project as an object

    Will raise an error if the project id cannot be found

    :param project_name: The name of the project

    :return: The id of the project
    :rtype: str

    :raises ValueError, ApiException

    :Examples:

    .. code-block:: python

        from wrapica.project import get_project_id_from_project_name

        project_obj = get_project_obj_from_project_name("my_project")

        print(project_obj.id)
        # "1234-5678-9012-3456"

    """
    try:
        return next(
            filter(
                lambda project_obj_iter: project_obj_iter.name == project_name,
                list_projects()
            )
        )
    except StopIteration:
        logger.error(f"Could not find project object from project name {project_name}")
        raise StopIteration


def get_project_id_from_project_name(
    project_name: str
) -> str:
    """
    Given a project name return the id of the project
    Will raise an error if the project id cannot be found

    :param project_name: The name of the project

    :return: The id of the project
    :rtype: str

    :raises StopIteration, ApiException

    :Examples:

    .. code-block:: python

        from wrapica.project import get_project_id_from_project_name

        project_id = get_project_id_from_project_name("my_project")

        print(project_id)
        # "1234-5678-9012-3456"
    """
    try:
        return next(filter(
            lambda kv_iter_: kv_iter_[1] == project_name,
            _get_project_mapping_dict().items()
        ))[0]
    except StopIteration as e:
        raise StopIteration("Could not find project id from project name: %s" % project_name) from e


# And vice-versa
def get_project_name_from_project_id(
        project_id: Union[UUID4, str]
) -> str:
    """
    Given a project id, get the project object and return the name attribute
    :param project_id:
    :return:
    """
    try:
        return next(filter(
            lambda kv_iter_: kv_iter_[0] == project_id,
            _get_project_mapping_dict().items()
        ))[1]
    except StopIteration as e:
        raise StopIteration("Could not find project name from project id: %s" % project_id) from e


def check_project_has_data_sharing_enabled(
        project_id: Union[UUID4, str]
) -> bool:
    """
    Given a project id return whether the project has data sharing enabled

    :param project_id: The id of the project

    :return: True if data sharing is enabled, False otherwise
    :rtype: str

    :raises ValueError, ApiException

    :Examples:

    .. code-block:: python

        from wrapica.project import get_project_id_from_project_name

        project_id = get_project_id_from_project_name("my_project")

        print(check_project_has_data_sharing_enabled(project_id))
        # False
    """

    # Configuration
    configuration = get_icav2_configuration()

    # Collect the projects
    # We assume that there aren't more than 1000 projects anyway
    # Enter a context with an instance of the API client
    with ApiClient(configuration) as api_client:
        api_instance = ProjectApi(api_client)

    # example passing only required values which don't have defaults set
    # and optional values
    try:
        # Retrieve a list of projects.
        api_response: Project = api_instance.get_project(str(project_id))
    except ApiException as e:
        logger.error("Exception when calling ProjectApi->get_project_by_id: %s\n" % e)
        raise ApiException

    return api_response.data_sharing_enabled


def list_projects(include_hidden_projects: bool = False) -> List[Project]:
    """
    List all projects

    :param include_hidden_projects:

    :return: List of project objects
    :rtype: List[`Project <https://umccr.github.io/libica/openapi/v3/docs/Project/>`_]

    :raises: ApiException

    :Examples:

    .. code-block:: python
        :linenos:

        from wrapica.project import list_projects
        all_active_projects = list_projects()
    """

    # Create api instance
    with ApiClient(get_icav2_configuration()) as api_client:
        api_instance = ProjectApi(api_client)

    # Set other parameters
    page_size = LIBICAV2_DEFAULT_PAGE_SIZE
    page_offset = 0

    # Initialise project list
    project_list = []

    # example passing only required values which don't have defaults set
    # and optional values
    while True:
        try:
            # Retrieve a list of projects.
            api_response = api_instance.get_projects(
                include_hidden_projects=include_hidden_projects,
                page_size=str(page_size),
                page_offset=str(page_offset)
            )
        except ApiException as e:
            logger.error("Exception when calling ProjectApi->get_projects: %s\n" % e)
            raise ApiException
        project_list.extend(api_response.items)

        # Check page offset and page size against total item count
        if page_offset + page_size >= api_response.total_item_count:
            break
        page_offset += page_size

    return project_list


def coerce_project_id_or_name_to_project_obj(project_id_or_name: str) -> Project:
    """
    Given a project id or name, coerce to a project object

    :param project_id_or_name: The project id or name

    :return: The project object
    """
    # Check if the input is in uuid format
    if is_uuid_format(project_id_or_name):
        return get_project_obj_from_project_id(project_id_or_name)

    return get_project_obj_from_project_name(project_id_or_name)


def coerce_project_id_or_name_to_project_id(project_id_or_name: str) -> str:
    """
    Given a project id or name, return the project id

    :param project_id_or_name: The id or name of the project

    :return: The id of the project
    :rtype: str

    :raises ValueError, ApiException

    :Examples:

    .. code-block:: python

        from wrapica.project import coerce_project_id_or_name_to_project_id

        project_id = coerce_project_id_or_name_to_project_id("my_project")

        print(project_id)
        # "1234-5678-9012-3456"
    """

    # Check if the input is in uuid format
    if is_uuid_format(project_id_or_name):
        return project_id_or_name

    return get_project_id_from_project_name(project_id_or_name)


def get_project_id() -> str:
    # Try get project id from env var
    try:
        project_id = get_project_id_from_env_var()
    except EnvironmentError:
        logger.debug("Could not get environment variable for project id, trying session file")
    else:
        return project_id

    # Try get project id from session file
    try:
        project_id = get_project_id_from_session_file()
    except KeyError:
        logger.error("Could not get project id from session file")
    else:
        return project_id

    # Could not get project id from either the session
    raise ValueError("Could not get project id from either environment variable or session file")



#!/usr/bin/env python

# Standard imports
from typing import Union

from libica.openapi.v3 import ProjectPipelineV4
# Libica models
from libica.openapi.v3.models import (
    ActivationCodeDetail,
    AnalysisInputDataMount,
    AnalysisInputExternalData,
    AnalysisV3,
    AnalysisV4,
    AnalysisStorageV3,
    AnalysisStorageV4,
    CreateCwlAnalysis,
    CreateNextflowAnalysis,
    CwlAnalysisJsonInput,
    CwlAnalysisStructuredInput,
    InputParameter,
    InputParameterList,
    NextflowAnalysisInput,
    PipelineConfigurationParameter,
    PipelineConfigurationParameterList,
    PipelineFile,
    Project,
    ProjectData,
    ProjectPipeline,
    SearchMatchingActivationCodesForCwlAnalysis,
    SearchMatchingActivationCodesForNextflowAnalysis
)

# Import everything
from .functions.project_pipelines_functions import (
    # Project Data functions
    get_project_pipeline_obj,
    get_project_pipeline_obj_from_pipeline_code,
    get_project_pipeline_id_from_pipeline_code,
    get_default_analysis_storage_obj_from_project_pipeline,
    get_default_analysis_storage_id_from_project_pipeline,
    get_project_pipeline_description_from_pipeline_id,
    get_analysis_storage_id_from_analysis_storage_size,
    coerce_pipeline_id_or_code_to_project_pipeline_obj,
    get_analysis_storage_from_analysis_storage_id,
    get_analysis_storage_from_analysis_storage_size,
    coerce_analysis_storage_id_or_size_to_analysis_storage,
    create_cwl_input_json_analysis_obj,
    launch_cwl_workflow,
    launch_nextflow_workflow,
    get_project_pipeline_input_parameters,
    get_project_pipeline_configuration_parameters,
    convert_icav2_uris_to_data_ids_from_cwl_input_json,
    convert_uris_to_data_ids_from_cwl_input_json,
    list_project_pipelines,
    is_pipeline_in_project,
    list_projects_with_pipeline,
    create_blank_params_xml,
    create_params_xml,
    release_project_pipeline,
    update_pipeline_file,
    delete_pipeline_file,
    add_pipeline_file,
    create_cwl_project_pipeline,
    create_cwl_workflow_from_zip,
    create_nextflow_pipeline_from_zip,
    create_nextflow_pipeline_from_nf_core_zip,
    create_nextflow_project_pipeline,
)

from .classes.analysis import (
    # Wrapica classes
    ICAv2AnalysisInput,
    ICAv2PipelineAnalysisTags,
    ICAv2EngineParameters
)

from .classes.cwl_analysis import (
    # Wrapica classes
    ICAv2CwlAnalysisJsonInput,
    ICAv2CWLEngineParameters,
    ICAv2CWLPipelineAnalysis
)

from .classes.nextflow_analysis import (
    # Wrapica classes
    ICAv2NextflowAnalysisInput,
    ICAv2NextflowEngineParameters,
    ICAv2NextflowPipelineAnalysis
)

AnalysisType = Union[AnalysisV3, AnalysisV4]
AnalysisStorageType = Union[AnalysisStorageV3, AnalysisStorageV4]
ProjectPipelineType = Union[ProjectPipeline, ProjectPipelineV4]

__all__ = [
    # Libica models
    'ActivationCodeDetail',
    'AnalysisInputDataMount',
    'AnalysisInputExternalData',
    'AnalysisStorageType',
    'CreateCwlAnalysis',
    'CreateNextflowAnalysis',
    'CwlAnalysisJsonInput',
    'CwlAnalysisStructuredInput',
    'InputParameter',
    'InputParameterList',
    'NextflowAnalysisInput',
    'PipelineConfigurationParameter',
    'PipelineConfigurationParameterList',
    'PipelineFile',
    'Project',
    'ProjectData',
    'ProjectPipeline',
    'SearchMatchingActivationCodesForCwlAnalysis',
    'SearchMatchingActivationCodesForNextflowAnalysis',
    # Type Unions
    'AnalysisType',
    'AnalysisStorageType',
    'ProjectPipelineType',
    # Functions
    'get_project_pipeline_obj',
    'get_project_pipeline_obj_from_pipeline_code',
    'get_project_pipeline_id_from_pipeline_code',
    'get_default_analysis_storage_obj_from_project_pipeline',
    'get_default_analysis_storage_id_from_project_pipeline',
    'get_project_pipeline_description_from_pipeline_id',
    'get_analysis_storage_id_from_analysis_storage_size',
    'coerce_pipeline_id_or_code_to_project_pipeline_obj',
    'get_analysis_storage_from_analysis_storage_id',
    'get_analysis_storage_from_analysis_storage_size',
    'coerce_analysis_storage_id_or_size_to_analysis_storage',
    'create_cwl_input_json_analysis_obj',
    'launch_cwl_workflow',
    'launch_nextflow_workflow',
    'get_project_pipeline_input_parameters',
    'get_project_pipeline_configuration_parameters',
    'convert_icav2_uris_to_data_ids_from_cwl_input_json',
    'convert_uris_to_data_ids_from_cwl_input_json',
    'list_project_pipelines',
    'is_pipeline_in_project',
    'list_projects_with_pipeline',
    'create_blank_params_xml',
    'create_params_xml',
    'release_project_pipeline',
    'update_pipeline_file',
    'delete_pipeline_file',
    'add_pipeline_file',
    'create_cwl_project_pipeline',
    'create_cwl_workflow_from_zip',
    'create_nextflow_pipeline_from_zip',
    'create_nextflow_pipeline_from_nf_core_zip',
    'create_nextflow_project_pipeline',
    # classes
    'ICAv2AnalysisInput',
    'ICAv2PipelineAnalysisTags',
    'ICAv2EngineParameters',
    'ICAv2CwlAnalysisJsonInput',
    'ICAv2CWLEngineParameters',
    'ICAv2CWLPipelineAnalysis',
    'ICAv2NextflowAnalysisInput',
    'ICAv2NextflowEngineParameters',
    'ICAv2NextflowPipelineAnalysis',
]
